# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
# AccessSlottedDevice
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QDialog, QMessageBox, QToolButton

from karabo.common.states import State
from karabo.native import AccessLevel, Configurable, Slot, String
from karabogui.binding.api import (
    DeviceProxy, PropertyProxy, apply_default_configuration, build_binding)
from karabogui.const import IS_MAC_SYSTEM
from karabogui.testing import (
    access_level, click_button, set_proxy_value, singletons)
from karabogui.topology.system_tree import SystemTreeNode

from ..command import DisplayCommand


class EmptyDevice(Configurable):
    """What a device looks like to a scene when it is offline"""


class SlottedDevice(Configurable):
    state = String(defaultValue=State.INIT)

    @Slot(allowedStates=[State.INIT],
          displayedName="Call ME", requiredAccessLevel=AccessLevel.OBSERVER)
    def callme(self):
        pass

    @Slot(allowedStates=[State.ACTIVE],
          requiredAccessLevel=AccessLevel.OBSERVER)
    def yep(self):
        pass


class AccessSlottedDevice(Configurable):
    state = String(defaultValue=State.INIT)

    @Slot(allowedStates=[State.INIT],
          displayedName="Call ME", requiredAccessLevel=AccessLevel.EXPERT)
    def callme(self):
        pass

    @Slot(allowedStates=[State.ACTIVE],
          requiredAccessLevel=AccessLevel.OPERATOR)
    def yep(self):
        pass


def get_slot_proxy(additional=False, klass=SlottedDevice):
    schema = klass.getClassSchema()
    binding = build_binding(schema)
    apply_default_configuration(binding)

    device = DeviceProxy(device_id="dev", server_id="Test", binding=binding)
    slot_proxy = PropertyProxy(root_proxy=device, path="callme")
    if additional:
        return slot_proxy, PropertyProxy(root_proxy=device, path="yep")
    else:
        return slot_proxy


def test_command_controller_basics(gui_app):
    slot_proxy = get_slot_proxy()
    controller = DisplayCommand(proxy=slot_proxy)
    controller.create(None)

    assert controller.widget is not None
    assert isinstance(controller._button, QToolButton)
    assert controller._actions[0].action.text() == "Call ME"
    assert controller._actions[0].action.toolTip() == "dev.callme"

    controller.destroy()
    assert controller.widget is None


def test_command_controller_access_level():
    slot_proxy, add_slot_proxy = get_slot_proxy(additional=True,
                                                klass=AccessSlottedDevice)
    controller = DisplayCommand(proxy=slot_proxy)
    controller.create(None)
    assert controller.visualize_additional_property(add_slot_proxy)

    assert controller.widget is not None
    assert isinstance(controller._button, QToolButton)
    assert controller._actions[0].action.text() == "Call ME"
    assert controller._actions[0].action.toolTip() == "dev.callme"
    assert controller._actions[1].action.text() == "yep"
    assert controller._actions[1].action.toolTip() == "dev.yep"

    with access_level(AccessLevel.OBSERVER):
        controller.setEnabled(None)
        assert not controller._actions[0].action.isEnabled()
        assert not controller._actions[1].action.isEnabled()

    with access_level(AccessLevel.EXPERT):
        controller.setEnabled(None)
        assert controller._actions[0].action.isEnabled()
        # Wrong state, correct access level
        assert not controller._actions[1].action.isEnabled()
        # set new state
        set_proxy_value(slot_proxy, "state", State.ACTIVE.value)
        assert controller._actions[1].action.isEnabled()

    with access_level(AccessLevel.OPERATOR):
        controller.setEnabled(None)
        # Actice state
        assert not controller._actions[0].action.isEnabled()
        assert controller._actions[1].action.isEnabled()

        set_proxy_value(slot_proxy, "state", State.INIT.value)
        assert not controller._actions[0].action.isEnabled()
        assert not controller._actions[1].action.isEnabled()


def test_command_controller_trigger(gui_app, mocker):
    slot_proxy = get_slot_proxy()
    controller = DisplayCommand(proxy=slot_proxy)
    controller.create(None)
    network = mocker.Mock()
    with singletons(network=network):
        controller._actions[0].action.trigger()
        network.onExecute.assert_called_with("dev", "callme", False)


def test_command_controller_macro_trigger(gui_app, mocker):
    slot_proxy = get_slot_proxy()
    node = SystemTreeNode()
    node.attributes = {"type": "macro"}
    slot_proxy.root_proxy.topology_node = node
    controller = DisplayCommand(proxy=slot_proxy)
    controller.create(None)
    network = mocker.Mock()
    with singletons(network=network):
        controller._actions[0].action.trigger()
        network.onExecute.assert_called_with("dev", "callme", True)


def test_command_controller_additional_proxy(gui_app):
    slot_p1, slot_p2 = get_slot_proxy(additional=True)
    controller = DisplayCommand(proxy=slot_p1)
    controller.create(None)

    assert controller.visualize_additional_property(slot_p2)
    # no displayedName, should be path
    assert controller._actions[1].action.text() == "yep"
    assert len(controller._actions) == 2
    assert slot_p2 in controller.proxies
    assert slot_p2.key in controller.model.keys

    assert controller.remove_additional_property(slot_p2)
    assert len(controller._actions) == 1
    assert slot_p2 not in controller.proxies
    assert slot_p2.key not in controller.model.keys

    # Already removed
    assert not controller.remove_additional_property(slot_p2)


def test_command_controller_state_change(gui_app):
    schema = SlottedDevice.getClassSchema()
    binding = build_binding(schema)
    apply_default_configuration(binding)

    dev_proxy = DeviceProxy(device_id="dev", server_id="swerver",
                            binding=binding)
    slot_proxy = PropertyProxy(root_proxy=dev_proxy, path="yep")
    controller = DisplayCommand(proxy=slot_proxy)
    controller.create(None)

    # device state is INIT
    assert not controller._actions[0].action.isEnabled()

    set_proxy_value(slot_proxy, "state", State.ACTIVE.value)
    # now `yep` should be enabled
    assert controller._actions[0].action.isEnabled()


def test_command_controller_button_finalization(gui_app):
    slot1, slot2 = get_slot_proxy(additional=True, klass=EmptyDevice)

    slotSchema = SlottedDevice()

    controller = DisplayCommand(proxy=slot1)
    controller.create(None)
    controller.visualize_additional_property(slot2)

    assert controller._actions[0].action.text() == "NO TEXT"
    assert controller._actions[1].action.text() == "NO TEXT"

    # The device "comes online"
    schema = slotSchema.getClassSchema()
    build_binding(schema, existing=slot1.root_proxy.binding)

    assert controller._actions[0].action.text() == "Call ME"
    assert controller._actions[1].action.text() == "yep"

    # The device "has schema injection"
    slotSchema.callme.descriptor.displayedName = "Injected"
    schema = slotSchema.getClassSchema()
    build_binding(schema, existing=slot1.root_proxy.binding)
    assert controller._actions[0].action.text() == "Injected"
    assert controller._actions[1].action.text() == "yep"


def test_command_controller_confirmation_dialog(gui_app, mocker):
    slot_proxy = get_slot_proxy()
    controller = DisplayCommand(proxy=slot_proxy)
    controller.create(None)
    network = mocker.Mock()
    widget = controller.widget

    with singletons(network=network):
        controller.model.requires_confirmation = False
        button = controller._button
        click_button(controller._button)

        assert not controller.model.requires_confirmation
        assert network.onExecute.call_count == 1
        assert not controller.widget.font().bold()

    with singletons(network=network):
        controller._requires_confirmation_slot()
        assert controller.model.requires_confirmation
        assert "bold" in widget.styleSheet()
        if not IS_MAC_SYSTEM:
            sh = ("QToolButton{font-size: 10pt; font: bold;}"
                  "QToolButton:enabled{color: rgb(255, 145, 255);}")
            assert sh in widget.styleSheet()

        # We don"t want to execute
        QMessageBox.question = mocker.MagicMock(return_value=QMessageBox.No)
        click_button(button)
        assert network.onExecute.call_count == 1
        # We want to execute
        QMessageBox.question = mocker.MagicMock(return_value=QMessageBox.Yes)
        click_button(button)
        assert network.onExecute.call_count == 2

    controller._requires_confirmation_slot()
    if not IS_MAC_SYSTEM:
        style = "QToolButton { font-size: 10pt; font: normal }"
        assert style in widget.styleSheet()

    QMessageBox.question = mocker.MagicMock(return_value=QMessageBox.No)

    format_action = controller.widget.actions()[1]
    path = "karabogui.controllers.display.command.FormatLabelDialog"
    dialog = mocker.patch(path)
    assert controller.model.font_size == 10
    if not IS_MAC_SYSTEM:
        style = "QToolButton { font-size: 10pt; font: normal }"
        assert style in widget.styleSheet()
    dialog().font_size = 14
    dialog().font_weight = "bold"
    dialog().exec.return_value = QDialog.Accepted
    format_action.trigger()
    assert controller.model.font_size == 14
    if not IS_MAC_SYSTEM:
        style = "QToolButton { font-size: 14pt; font: bold }"
        assert style in widget.styleSheet()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
# NOTE: None of the controller classes are here. They should be acquired from
# the registry, not directly (tests excluded, of course)
from .arrays import (
    DIMENSIONS, get_array_data, get_dimensions_and_encoding, get_image_data)
from .base import BaseBindingController
from .basearray import BaseArrayGraph
from .baselabel import AlarmMixin, BaseLabelController, FormatMixin
from .baselineedit import BaseLineEditController
from .registry import (
    get_class_const_trait, get_compatible_controllers, get_model_controller,
    get_scene_model_class, register_binding_controller)
from .trendmodel import (
    HIDDEN, ONE_DAY, ONE_HOUR, ONE_WEEK, TEN_MINUTES, UPTIME, Curve,
    get_start_end_date_time)
from .unitlabel import add_unit_label
from .util import (
    axis_label, get_class_const_trait, get_regex, has_options, has_regex,
    is_proxy_allowed, populate_controller_registry, with_display_type)
from .validators import BindingValidator, ListValidator, SimpleValidator



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on October 9, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op

from qtpy.QtWidgets import QAction
from traits.api import Instance, on_trait_change

from karabo.common.api import State
from karabo.common.scenemodel.api import ColorBoolModel
from karabogui import icons
from karabogui.binding.api import BoolBinding, get_binding_value
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.icons.statefulicons.color_change_icon import (
    ColorChangeIcon, get_color_change_icon)
from karabogui.indicators import STATE_COLORS
from karabogui.widgets.hints import SvgWidget


@register_binding_controller(ui_name='Switch Bool',
                             klassname='DisplayColorBool',
                             binding_type=BoolBinding)
class DisplayColorBool(BaseBindingController):
    # The scene data model class for this controller
    model = Instance(ColorBoolModel, args=())
    # A icon which will be used in different colors
    icon = Instance(ColorChangeIcon)

    def _icon_default(self):
        path = op.join(op.dirname(icons.__file__), 'switch-bool.svg')
        return get_color_change_icon(path)

    def create_widget(self, parent):
        widget = SvgWidget(parent)
        logicAction = QAction("Invert color logic", widget)
        logicAction.triggered.connect(self.logic_action)
        widget.addAction(logicAction)
        # update the context menu and keep track
        logicAction.setCheckable(True)
        logicAction.setChecked(self.model.invert)

        return widget

    def value_update(self, proxy):
        value = get_binding_value(proxy)
        if self.widget is None or value is None:
            # We might get here when the scene model changes and the widget
            # is not yet ready or the property proxy is not set.
            return

        if not self.model.invert:
            color_state = State.ACTIVE if value else State.PASSIVE
        else:
            color_state = State.PASSIVE if value else State.ACTIVE
        # Set the tooltip according to the value
        self.widget.setToolTip(f"{value}")
        svg = self.icon.with_color(STATE_COLORS[color_state])
        self.widget.load(bytearray(svg, encoding='UTF-8'))

    @on_trait_change('model.invert')
    def _invert_update(self):
        if self.proxy is not None:
            self.value_update(self.proxy)

    def logic_action(self):
        self.model.invert = not self.model.invert



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtWidgets import QBoxLayout

from karabo.common.scenemodel.api import BoxLayoutModel, LabelModel, SceneModel
from karabogui.const import IS_MAC_SYSTEM

from ...view import SceneView
from ..actions import BoxHSceneAction, BoxVSceneAction


@pytest.fixture()
def scene_model(gui_app):
    foo_label = LabelModel(x=20, y=20, text="foo")
    bar_label = LabelModel(x=10, y=10, text="bar")
    baz_label = LabelModel(x=10, y=20, text="baz")
    labels = [foo_label, bar_label, baz_label]
    scene_model = SceneModel(children=labels)
    return scene_model


@pytest.fixture()
def scene_view(gui_app, scene_model):
    scene_view = SceneView(model=scene_model)
    return scene_view


def test_basics(scene_view, scene_model):
    _assert_sceneview_contents(scene_view, scene_model, changed=False)


@pytest.mark.skipif(IS_MAC_SYSTEM, reason="Segfault on MacOS")
def test_horz_layout_action(scene_view, scene_model):
    action = BoxHSceneAction()

    # Not selected, so the values are not changed
    action.perform(scene_view)
    _assert_sceneview_contents(scene_view, scene_model, changed=False)
    _assert_layout_in_scene(scene_view, model_klass=BoxLayoutModel,
                            valid=False)

    # Select all label models
    scene_view.select_all()
    action.perform(scene_view)
    _assert_sceneview_contents(scene_view, scene_model, changed=True)
    _assert_layout_in_scene(scene_view, model_klass=BoxLayoutModel,
                            valid=True)

    # Check layout model
    layout_model = _get_layout_model(scene_view, model_klass=BoxLayoutModel)
    assert layout_model.direction == QBoxLayout.LeftToRight

    # Check order. The expected is as follows: (x, y)
    # 1. The label bar goes first (10, 10)
    # 2. The label baz goes next (10, 20)
    # 3, The label foo goes last (20, 20)
    labels = ["bar", "baz", "foo"]
    texts = [child.text for child in layout_model.children]
    assert labels == texts
    scene_view.destroy()


@pytest.mark.skipif(IS_MAC_SYSTEM, reason="Segfault on MacOS")
def test_vert_layout_action(scene_view, scene_model):
    action = BoxVSceneAction()

    # Not selected, so the values are not changed
    action.perform(scene_view)
    _assert_sceneview_contents(scene_view, scene_model, changed=False)
    _assert_layout_in_scene(scene_view, model_klass=BoxLayoutModel,
                            valid=False)

    # Select all label models
    scene_view.select_all()
    action.perform(scene_view)
    _assert_sceneview_contents(scene_view, scene_model, changed=True)
    _assert_layout_in_scene(scene_view, model_klass=BoxLayoutModel, valid=True)

    # Check layout model
    layout_model = _get_layout_model(scene_view, model_klass=BoxLayoutModel)
    assert layout_model.direction == QBoxLayout.TopToBottom

    # Check order. The expected is as follows: (x, y)
    # 1. The label bar goes first (10, 10)
    # 2. The label baz goes next (10, 20)
    # 3, The label foo goes last (20, 20)
    labels = ["bar", "baz", "foo"]
    texts = [child.text for child in layout_model.children]
    assert labels == texts
    scene_view.destroy()


def _assert_sceneview_contents(view, model, changed=True):
    assertion = not changed
    obj_cache = list(view._scene_obj_cache.keys())
    assert (model.children == obj_cache) == assertion


def _assert_layout_in_scene(view, model_klass=None, valid=True):
    layout_model = _get_layout_model(view, model_klass)
    if valid:
        assert layout_model is not None
    else:
        assert layout_model is None


def _get_layout_model(view, model_klass=None):
    # Get layout model from scene objects
    for model in view._scene_obj_cache.keys():
        if isinstance(model, model_klass):
            # We bail out immediately since the obj cache respects order
            return model



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 10, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from traits.api import Instance, Int

from karabo.common.scenemodel.api import LineEditModel
from karabogui.binding.api import (
    CharBinding, StringBinding, VectorCharBinding, get_editor_value)
from karabogui.controllers.api import (
    BaseBindingController, is_proxy_allowed, register_binding_controller)
from karabogui.util import SignalBlocker
from karabogui.widgets.hints import LineEdit

BINDING_TYPES = (CharBinding, StringBinding, VectorCharBinding)


@register_binding_controller(ui_name='Text Field', can_edit=True,
                             klassname='EditableLineEdit', priority=10,
                             binding_type=BINDING_TYPES)
class EditableLineEdit(BaseBindingController):
    model = Instance(LineEditModel, args=())
    # Internal details
    _last_cursor_pos = Int(0)

    def create_widget(self, parent):
        widget = LineEdit(parent)
        widget.textChanged.connect(self._on_text_changed)
        return widget

    def binding_update(self, proxy):
        if isinstance(proxy.binding, CharBinding):
            mask = 'X'
            with SignalBlocker(self.widget):
                self.widget.setInputMask(mask)

    def value_update(self, proxy):
        value = get_editor_value(proxy, '')
        if not isinstance(value, str):
            value = value.decode()

        with SignalBlocker(self.widget):
            self.widget.setText(value)
        self.widget.setCursorPosition(self._last_cursor_pos)

    def _on_text_changed(self, value):
        if self.proxy.binding is None:
            return
        self._last_cursor_pos = self.widget.cursorPosition()
        self.proxy.edit_value = value

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self.widget.setEnabled(enable)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialog

from karabogui.validators import NumberValidator

DECIMALS = 6


class RangeDialog(QDialog):
    def __init__(self, config=None, actual=None, axis=0, parent=None):
        super().__init__(parent)
        self.setModal(False)
        ui_path = Path(__file__).parent.joinpath("range_config.ui")
        uic.loadUi(ui_path, self)
        self.get_axis_notation(axis=axis)
        self.ui_min.setValidator(NumberValidator(decimals=DECIMALS))
        self.ui_max.setValidator(NumberValidator(decimals=DECIMALS))

        # Show the actual view!
        self.ui_view_min.setText(f"{actual[self.axis_min]:.2e}")
        self.ui_view_max.setText(f"{actual[self.axis_max]:.2e}")

        # Load the model settings!
        self.ui_min.setText(f"{config[self.axis_min]}")
        self.ui_max.setText(f"{config[self.axis_max]}")

        state = config[self.axis_auto]
        self.ui_autorange.setChecked(state)
        self.ui_autorange.toggled.connect(self._check_box_triggered)
        self.ui_min.setEnabled(not state)
        self.ui_max.setEnabled(not state)
        if not state:
            self.ui_min.setFocus()
            self.ui_min.selectAll()

    def get_axis_notation(self, axis):
        axis = "x" if axis == 0 else "y"
        self.axis_min = f"{axis}_min"
        self.axis_max = f"{axis}_max"
        self.axis_auto = f"{axis}_autorange"
        self.ui_axis.setText(f"Axis View: {axis}-axis")

    @Slot(bool)
    def _check_box_triggered(self, state):
        self.ui_min.setEnabled(not state)
        self.ui_max.setEnabled(not state)
        if not state:
            self.ui_min.setFocus()
            self.ui_min.selectAll()

    @property
    def limits(self):
        limits = [float(self.ui_min.text()), float(self.ui_max.text())]
        axis_min = min(limits)
        axis_max = max(limits)
        config = {
            self.axis_min: axis_min,
            self.axis_max: axis_max,
            self.axis_auto: self.ui_autorange.isChecked()}

        return config

    @staticmethod
    def get(model, actual, axis=0, parent=None):
        dialog = RangeDialog(model, actual, axis, parent)
        result = dialog.exec() == QDialog.Accepted
        content = {}
        content.update(dialog.limits)

        return content, result



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.scenemodel.api import TrendGraphModel
from karabo.native import Configurable, Float, Hash
from karabogui.binding.api import ProxyStatus, build_binding
from karabogui.generic_scenes import get_property_proxy_model
from karabogui.panels.widgetcontrollerpanel import WidgetControllerPanel
from karabogui.singletons.mediator import Mediator
from karabogui.testing import (
    get_property_proxy, set_proxy_hash, singletons, system_hash)
from karabogui.topology.api import SystemTopology


class Object(Configurable):
    prop = Float(
        defaultValue=0.2)


def test_panel_controller_basics(gui_app):
    """Test the basics of the widget controller"""
    mediator = Mediator()
    topology = SystemTopology()
    h = system_hash()
    topology.initialize(h)
    with singletons(mediator=mediator, topology=topology):
        proxy = get_property_proxy(Object.getClassSchema(), "prop",
                                   device_id="divvy")
        model = get_property_proxy_model(proxy)
        assert isinstance(model, TrendGraphModel)

        # Create a panel and start monitoring
        panel = WidgetControllerPanel("TestTitle", model)
        assert panel.controller._showing
        assert panel.proxy is not proxy
        proxy = panel.proxy
        # Root device is offline, we are showing the pixmap
        assert not panel.status_symbol.isHidden()
        assert proxy.visible
        # Root devices comes online
        proxy.root_proxy.status = ProxyStatus.ONLINE
        # Moves to requested
        assert proxy.root_proxy.status is ProxyStatus.ONLINEREQUESTED
        assert proxy.root_proxy._monitor_count == 1

        # Simulate schema and config walk through
        build_binding(Object.getClassSchema(), proxy.root_proxy.binding)
        assert proxy.root_proxy.status is ProxyStatus.SCHEMA

        set_proxy_hash(proxy, Hash("prop", 1.2))
        proxy.root_proxy.config_update = True

        assert proxy.root_proxy.status is ProxyStatus.MONITORING
        assert panel.status_symbol.isHidden()
        panel.close()

        # And we stop monitoring if we close the panel
        assert proxy.root_proxy._monitor_count == 0
        assert proxy.root_proxy.status is ProxyStatus.ONLINE
        assert not panel.controller._showing

        panel.destroy()


def test_repr(gui_app):
    """Test the basics of the widget controller"""
    mediator = Mediator()
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(mediator=mediator, topology=topology):
        proxy = get_property_proxy(Object.getClassSchema(), "prop",
                                   device_id="divvy")
        model = get_property_proxy_model(proxy)

        panel = WidgetControllerPanel("TestTitle", model)
        assert repr(panel) == "<WidgetControllerPanel property=divvy.prop>"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import ViewBox
from qtpy.QtCore import Qt

from karabogui.graph.common.api import KaraboViewBox, MouseTool


class KaraboImageViewBox(KaraboViewBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setBackgroundColor(None)

    # ---------------------------------------------------------------------
    # mouse events

    def mouseDragEvent(self, event, axis=None):
        if (self.mouse_mode is MouseTool.Picker
                and event.buttons() == Qt.LeftButton):
            event.ignore()
        else:
            super().mouseDragEvent(event, axis)

    # ---------------------------------------------------------------------
    # Public methods

    def set_mouse_tool(self, mode):
        if mode is MouseTool.Picker:
            vb_mode = ViewBox.RectMode
            cursor = Qt.PointingHandCursor
        else:
            super().set_mouse_tool(mode)
            return

        self.setMouseMode(vb_mode)
        self.setCursor(cursor)
        self.mouse_mode = mode



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on July 6, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from enum import Enum


class ProxyStatus(Enum):
    # device could, but is not started
    OFFLINE = "offline"
    # the device is online but doesn't have a schema yet
    ONLINE = "online"
    # online device waiting for its schema
    ONLINEREQUESTED = "onlinerequested"
    # everything is up-and-running
    ALIVE = "alive"
    # we are registered to monitor this device
    MONITORING = "monitoring"
    # a schema is requested, but didnt arrive yet
    REQUESTED = "requested"
    # the device has a schema, but no value yet
    SCHEMA = "schema"
    # device server not available
    NOSERVER = "noserver"
    # class plugin not available
    NOPLUGIN = "noplugin"
    # device running, but of different type
    INCOMPATIBLE = "incompatible"
    MISSING = "missing"


# The device is online in these statuses
ONLINE_STATUSES = (
    ProxyStatus.ONLINE,
    ProxyStatus.ALIVE,
    ProxyStatus.ONLINEREQUESTED,
    ProxyStatus.MONITORING,
    ProxyStatus.SCHEMA,
)
SCHEMA_STATUSES = (
    ProxyStatus.ALIVE,
    ProxyStatus.MONITORING,
    ProxyStatus.SCHEMA,
)
# The device is not able to receive configuration
NO_CONFIG_STATUSES = (
    ProxyStatus.NOPLUGIN,
    ProxyStatus.NOSERVER,
    ProxyStatus.REQUESTED,
)
NO_CLASS_STATUSES = (
    ProxyStatus.NOPLUGIN,
    ProxyStatus.NOSERVER
)

# The device can perform an online to offline configuration
ONLINE_CONFIG_STATUSES = (
    ProxyStatus.MONITORING,
    ProxyStatus.ALIVE,
)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from traits.api import Instance

from karabo.common.scenemodel.api import VectorGraphModel
from karabo.common.scenemodel.widgets.graph_plots import NDArrayGraphModel
from karabogui.binding.api import VectorNumberBinding
from karabogui.binding.binding_types import NDArrayBinding
from karabogui.controllers.api import register_binding_controller
from karabogui.controllers.basearray import BaseArrayGraph


@register_binding_controller(ui_name='Vector Graph', klassname='VectorGraph',
                             binding_type=VectorNumberBinding,
                             priority=0,
                             can_show_nothing=False)
class DisplayVectorGraph(BaseArrayGraph):
    """The VectorGraph controller for display basic vector data
    """
    model = Instance(VectorGraphModel, args=())


@register_binding_controller(ui_name='NDArray Graph',
                             klassname='NDArrayGraph',
                             binding_type=NDArrayBinding,
                             priority=90,
                             can_show_nothing=False)
class DisplayNDArrayGraph(BaseArrayGraph):
    """The NDArray controller for display of pulse data"""
    model = Instance(NDArrayGraphModel, args=())



#############################################################################
# Author: <steffen.hauf@xfel.eu>
# Created on December 7, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Qt
from qtpy.QtGui import QIcon, QPixmap, QStandardItem, QStandardItemModel
from qtpy.QtSvg import QSvgWidget
from qtpy.QtWidgets import QDialog, QListView
from traits.api import Instance

from karabo.common.scenemodel.api import StatefulIconWidgetModel
from karabogui.binding.api import StringBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller, with_display_type)
from karabogui.icons.statefulicons import ICONS
from karabogui.icons.statefulicons.color_change_icon import ColorChangeIcon
from karabogui.indicators import get_state_color

WHITE = '#ffffff'


@register_binding_controller(ui_name='Standard Icon',
                             klassname='StatefulIconWidget',
                             binding_type=StringBinding,
                             is_compatible=with_display_type('State'),
                             can_show_nothing=False)
class StatefulIconWidget(BaseBindingController):
    # The scene model class used by this controller
    model = Instance(StatefulIconWidgetModel, args=())
    # Internal traits
    _icon = Instance(ColorChangeIcon)

    def create_widget(self, parent):
        # show a list to pick the icon from if it is not set
        widget = QSvgWidget(parent)
        if not self.model.icon_name:
            # This is the creation phase, and the picker will set the icon
            # on the model
            self._show_icon_picker(widget)
        else:
            # Note: We might have an icon that is not supported. In this case
            # we do not modify the traits model!
            icon = ICONS.get(self.model.icon_name, ICONS["icon_default"])
            self._icon = icon

        return widget

    def value_update(self, proxy):
        if self._icon is None:
            return  # We haven't been configured correctly!
        value = proxy.value
        color = get_state_color(value)

        svg = self._icon.with_color(color)
        self.widget.load(bytearray(svg, encoding='UTF-8'))

    def _show_icon_picker(self, widget):
        dialog = QDialog(widget)
        dialog.setWindowTitle("Select standard icon")
        # we do not allow to cancel this dialog. An icon must be selected
        dialog.setWindowFlags(Qt.Window |
                              Qt.WindowTitleHint |
                              Qt.CustomizeWindowHint)

        iconlist = QListView(dialog)
        iconlist.setMinimumSize(400, 600)
        model = QStandardItemModel(iconlist)

        non_selection = ICONS["icon_default"]
        # we add all items with a preview icon
        for key, icon in sorted(ICONS.items()):
            listItem = QStandardItem(icon.description)
            p = QPixmap()
            p.loadFromData(bytearray(icon.with_color(WHITE), encoding='UTF-8'))
            listItem.setData(QIcon(p),
                             Qt.DecorationRole)
            listItem.setData(icon, Qt.UserRole + 1)
            model.appendRow(listItem)
        iconlist.setModel(model)

        # double clicking an entry will select it and close the dialog
        def handleDoubleClick(index):
            self._icon = index.data(Qt.UserRole + 1)
            self.model.icon_name = self._icon.name
            dialog.close()

        iconlist.doubleClicked.connect(handleDoubleClick)
        dialog.exec()
        if self._icon is None:
            # Operator bailed out of the dialog! Set a default icon!
            self._icon = non_selection
            self.model.icon_name = self._icon.name



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QWidget

from karabo.common.scenemodel.api import GlobalAlarmModel
from karabo.native import Configurable, String
from karabogui.indicators import get_alarm_svg
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..alarm import DisplayAlarm


class MockQSvgWidget(QWidget):
    def load(self, svg):
        self.loaded_data = svg


class AlarmObject(Configurable):
    alarmCondition = String(
        defaultValue="alarm",
        displayType="AlarmCondition")

    noAlarmCondition = String(
        defaultValue="alarm")

    anotherAlarmCondition = String(
        defaultValue="alarm",
        displayType="AlarmCondition")


schema = AlarmObject.getClassSchema()
new_schema = AlarmObject.getClassSchema()


def test_alarm_controller(gui_app, mocker):
    proxy = get_class_property_proxy(schema, "alarmCondition")
    new_proxy = get_class_property_proxy(new_schema, "alarmCondition")
    model = GlobalAlarmModel()

    target = "karabogui.controllers.display.alarm.QSvgWidget"

    # Patch to fake Svg behavior
    mocker.patch(target, new=MockQSvgWidget)

    proxy.value = "warn"
    controller = DisplayAlarm(proxy=proxy, model=model)
    controller.create(None)
    assert controller.widget is not None

    for alarm_type in ("warn", "interlock", "alarm", "none"):
        set_proxy_value(proxy, "alarmCondition", alarm_type)
        active = get_alarm_svg(alarm_type)
        assert controller.widget.loaded_data == active

    none_svg = get_alarm_svg("none")
    assert controller.widget.loaded_data == none_svg

    set_proxy_value(new_proxy, "alarmCondition", "warn")
    # Add a new proxy
    controller.add_proxy(new_proxy)
    # Not taken into account
    warn_svg = get_alarm_svg("warn")
    assert controller.widget.loaded_data != warn_svg

    # Change the old ``none`` alarm to ``alarm``
    set_proxy_value(proxy, "alarmCondition", "alarm")
    alarm_svg = get_alarm_svg("alarm")
    assert controller.widget.loaded_data == alarm_svg

    controller.clear_widget()
    assert controller.widget.loaded_data not in {alarm_svg, warn_svg, none_svg}

    controller.destroy()
    assert controller.widget is None



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 29, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QDialog, QMessageBox

from karabo.common.api import (
    KARABO_PROJECT_MANAGER, set_modified_flag, walk_traits_object)
from karabo.common.project.api import (
    BaseProjectObjectModel, DeviceConfigurationModel, DeviceInstanceModel,
    DeviceServerModel, MacroModel, ProjectModel, device_config_exists,
    device_instance_exists, macro_exists, read_lazy_object,
    recursive_save_object)
from karabo.common.sanity_check import validate_macro
from karabo.common.scenemodel.api import (
    BaseWidgetObjectData, SceneLinkModel, SceneModel)
from karabo.native import Hash, read_project_model
from karabogui import messagebox
from karabogui.access import AccessRole, access_role_allowed
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.request import onShutdown
from karabogui.singletons.api import (
    get_config, get_db_conn, get_network, get_project_model, get_topology)
from karabogui.topology.util import get_macro_servers
from karabogui.util import (
    create_list_string, move_to_cursor, send_info, version_compatible)


def add_device_to_server(server, class_id='', parent=None):
    """ Add a new device to a server

    :param server: A DeviceServerModel object, OR string containing a server id
    :param class_id: An optional parameter containing the class_id to use for
                     the newly created device.
    :return: The device id of the newly added device, or an empty string if
             no device was added.
    """
    from karabogui.project.dialog.device_handle import DeviceHandleDialog

    server_model = None
    if isinstance(server, str):
        server_model = get_device_server_model(server)
    elif isinstance(server, DeviceServerModel):
        server_model = server

    if server_model is None:
        msg = ('A device server with the ID "<b>{}</b>"<br>'
               'needs to be added to the current project before<br>'
               'a device can be added').format(server)
        messagebox.show_warning(msg, title='Server doesn\'t exist',
                                parent=parent)
        return ''

    dialog = DeviceHandleDialog(server_id=server_model.server_id,
                                class_id=class_id,
                                parent=parent)
    result = dialog.exec()
    if result == QDialog.Accepted:
        # Check for existing device
        if check_device_instance_exists(dialog.instance_id):
            return ''

        config_model = DeviceConfigurationModel(
            class_id=dialog.class_id, configuration=Hash(),
            simple_name=dialog.configuration_name,
        )
        # Set initialized and modified last
        config_model.initialized = config_model.modified = True

        device = DeviceInstanceModel(
            class_id=dialog.class_id,
            instance_id=dialog.instance_id,
            configs=[config_model],
            active_config_ref=config_model.uuid,
        )
        # Set initialized and modified last
        device.initialized = device.modified = True

        server_model.devices.append(device)
        return device.instance_id

    # No device was added!
    return ''


def check_device_instance_exists(instance_id):
    """Check whether ``instance_id`` already exists in the current projects

    Shows a messagebox if the device already exists in the project!

    :returns: ``True`` if ``instance_id`` was NOT found else ``False``.
    """
    root_project = get_project_model().root_model
    if device_instance_exists(root_project, instance_id):
        msg = ('Another device with the same device ID \"<b>{}</b>\" '
               '<br>already exists! Therefore it will not be '
               'added!').format(instance_id)
        messagebox.show_warning(msg, title='Device already exists')
        return True
    return False


def validate_device_instance_exists(instance_id):
    """Validate whether ``instance_id`` was found in the root project

        Shows a messagebox if the device was not found in the project!

        :returns: ``True`` if ``instance_id`` was found else ``False``.
    """
    root_project = get_project_model().root_model
    if not device_instance_exists(root_project, instance_id):
        msg = ('The device with the device ID \"<b>{}</b>\" '
               '<br> was not found in the project!').format(instance_id)
        messagebox.show_warning(msg, title='Device does not exist')
        return False
    return True


def check_device_config_exists(instance_id, config_name):
    """Check whether the configuration already exists for the device

    :param instance_id: deviceId to be searched for
    :param config_name: device configuration to be searched for

    :return ``True`` if config name is found else ``False``.
    """
    root_project = get_project_model().root_model
    if device_config_exists(root_project, instance_id, config_name):
        msg = ('Another device configuration with the same name \"<b>{}</b>\" '
               '<br>already exists! Therefore it will not be '
               'added!').format(config_name)
        messagebox.show_warning(msg,
                                title='Device configuration already exists')
        return True
    return False


def check_macro_exists(instance_id):
    """Check whether the macro name already exists

    :param instance_id: serverId to be searched for

    :return ``True`` if server name is found else ``False``.
    """
    root_project = get_project_model().root_model
    if macro_exists(root_project, instance_id):
        msg = ('The macro with the name \"<b>{}</b>\" '
               '<br> already exists. Therefore it will not be '
               'added!').format(instance_id)
        messagebox.show_warning(msg, title='Macro already exists')
        return True
    return False


def get_device_server_model(server_id):
    """Given a string containing a server id, return the corresponding
    ``DeviceServerModel`` object in the current project, or None if one does
    not exist.
    """
    server_model = None

    def visitor(model):
        nonlocal server_model
        if (isinstance(model, DeviceServerModel) and
                model.server_id == server_id):
            server_model = model
            return True

    root_project = get_project_model().root_model
    walk_traits_object(root_project, visitor, fast_exit=True)
    return server_model


def load_project(is_subproject=False, parent=None):
    """Load a project from the project database.

    :param is_subproject States whether a master or a subproject is about to be
                         loaded
    """
    from karabogui.project.dialog.project_handle import LoadProjectDialog
    dialog = LoadProjectDialog(is_subproject=is_subproject, parent=parent)
    result = dialog.exec()
    if result == QDialog.Accepted:
        domain, uuid = dialog.selected_item()
        if domain is not None and uuid is not None:
            db_conn = get_db_conn()
            model = ProjectModel(uuid=uuid)
            read_lazy_object(domain, uuid, db_conn, read_project_model,
                             existing=model)
            db_conn.default_domain = domain
            db_conn.flush()
            if not dialog.ignore_cache:
                # Set modified flag recursively to True to make sure that
                # EVERYTHING gets saved to the database in case the project was
                # loaded from cache
                set_modified_flag(model, value=True)
            if not is_subproject:
                broadcast_event(KaraboEvent.ProjectName,
                                {"simple_name": dialog.simple_name})
                send_info(type="open_project", uuid=model.uuid,
                          domain=domain, simple_name=dialog.simple_name)
            return model
    return None


def load_project_with_item(parent=None):
    """Load a project with a device specified by the user from the project
    database.
    """
    from karabogui.project.dialog.load_project_with_item import (
        LoadProjectWithDialog)
    domain = get_config()["domain"]
    dialog = LoadProjectWithDialog(
        domain=domain, parent=parent)
    result = dialog.exec()
    if result == QDialog.Accepted:
        domain = dialog.selected_domain()
        uuid = dialog.selected_project_id()
        get_config()["domain"] = domain
        if domain is not None and uuid is not None:
            db_conn = get_db_conn()
            model = ProjectModel(uuid=uuid)
            read_lazy_object(domain, uuid, db_conn, read_project_model,
                             existing=model)
            db_conn.default_domain = domain
            db_conn.flush()
            # Project name will be updated
            broadcast_event(KaraboEvent.ProjectName,
                            {"simple_name": dialog.simple_name})

            send_info(type="open_project_with_device", uuid=model.uuid,
                      domain=domain,
                      simple_name=dialog.simple_name)
            return model
    return None


def reload_project(model):
    """Reload a project from the project database."""
    uuid = model.uuid
    model = ProjectModel(uuid=uuid)
    # NOTE: We keep the default domain of the db_conn here!
    db_conn = get_db_conn()
    domain = db_conn.default_domain
    read_lazy_object(domain, uuid, db_conn, read_project_model,
                     existing=model)
    # NOTE: Cache objects are loaded without modified flag!
    db_conn.flush()

    return model


def maybe_save_modified_project(project, parent=None):
    """Check modified flag of the ``project`` and offer saving via dialog
    """
    if project is None:
        return True

    if show_save_project_message(project, parent):
        save_object(project)

    return True


def save_as_object(obj):
    """ Save copy of individual `obj` recursively into the project database

    :param obj A project model object
    """
    from karabogui.project.dialog.project_handle import NewProjectDialog

    # Map old scene UUIDs to new UUIDs
    scene_uuids = {}
    # Map old macro UUIDs to new UUIDs
    macro_uuids = {}
    # Store all scene link references
    scene_links = []
    # Store all models that link to macros
    macro_models = []

    def _visitor(model):
        if isinstance(model, BaseProjectObjectModel):
            old_uuid = model.uuid
            model.reset_uuid()
            if isinstance(model, SceneModel):
                # Keep track of new UUIDs
                scene_uuids[old_uuid] = model.uuid
            if isinstance(model, MacroModel):
                # Keep track of new UUIDs
                macro_uuids[old_uuid] = model.uuid
        elif isinstance(model, SceneLinkModel):
            scene_links.append(model)
        elif isinstance(model, BaseWidgetObjectData):
            for key in model.keys:
                # optimize out non macros
                if key.startswith("Macro-"):
                    # key will be => Macro-module-UUID-Classname.keyname
                    uuid = '-'.join(key.split('-')[2:-1])
                    macro_models.append((model, uuid))

    def _replace_scene_link_uuids():
        for link in scene_links:
            parts = link.target.split(':')
            if len(parts) != 2:
                break
            # target format => "simple_name:UUID"
            simple_name = parts[0]
            old_uuid = parts[1]
            if old_uuid in scene_uuids:
                new_uuid = scene_uuids[old_uuid]
                target = f"{simple_name}:{new_uuid}"
            else:
                target = f"{simple_name}:{old_uuid}"
            link.target = target

    def _replace_macro_uuids():
        for widget, old_uuid in macro_models:
            new_keys = []
            for key in widget.keys:
                if old_uuid in key:
                    # replace old uuid with new uuid
                    new_uuid = macro_uuids[old_uuid]
                    key = key.replace(old_uuid, new_uuid)
                new_keys.append(key)
            widget.keys = new_keys

    assert isinstance(obj, ProjectModel)
    dialog = NewProjectDialog(model=obj)
    if dialog.exec() == QDialog.Accepted:
        if obj.simple_name == dialog.simple_name:
            # Throw a tantrum
            messagebox.show_error(
                "Please choose a different name. "
                "Projects with the same name should be avoided.")
            return
        obj.simple_name = dialog.simple_name
        # Reset UUIDs and map scene and macro models
        walk_traits_object(obj, _visitor)
        # Replace old scene link UUIDs with new ones
        _replace_scene_link_uuids()
        # Replace old macro UUIDs keys with new ones
        _replace_macro_uuids()
        # Set all child object of the given ``obj`` to modified to actually
        # save the complete tree to the new domain
        set_modified_flag(obj, value=True)
        save_object(obj, dialog.domain)


def save_object(obj, domain=None):
    """ Save individual `obj` recursively into the project database

    :param obj A project model object
    """
    if not _is_karabo3():
        dialog = QMessageBox(QMessageBox.Critical, "Error",
                             "Can not Save project to Karabo2")
        move_to_cursor(dialog)
        dialog.exec()
        return
    db_conn = get_db_conn()
    if domain is None:
        domain = db_conn.default_domain

    recursive_save_object(obj, db_conn, domain)
    db_conn.default_domain = domain
    db_conn.flush()


def show_save_project_message(project, parent=None):
    """Check whether the given ``project`` is modified and show a messagebox to
    allow the user to confirm saving or cancel

    :return Whether the user wants to save
    """

    if (project is not None and project.modified
            and access_role_allowed(AccessRole.PROJECT_EDIT)):
        ask = ('The project \"<b>{}</b>\" has been modified.<br />Do you want '
               'to save the project?').format(project.simple_name)
        options = (QMessageBox.Save | QMessageBox.No)
        reply = QMessageBox.question(parent, 'Save project', ask, options,
                                     QMessageBox.No)
        if reply == QMessageBox.No:
            return False

        if reply == QMessageBox.Save:
            return True
    return False


def show_modified_project_message(project, parent=None):
    """Check whether the given ``project`` is modified and show a messagebox to
    allow the user to confirm saving or cancel

    :return Whether the user wants to save
    """
    if project is None:
        return False

    if project.modified:
        ask = ('The project \"<b>{}</b>\" has been modified.<br />Are you '
               'sure you want to continue? <b>Your project changes will be '
               'lost</b>!').format(project.simple_name)
        options = (QMessageBox.Yes | QMessageBox.No)
        reply = QMessageBox.question(parent, 'Modified project', ask, options,
                                     QMessageBox.No)
        if reply == QMessageBox.No:
            return False

        if reply == QMessageBox.Yes:
            return True
    return True


def show_reload_scene_project_message(data, parent=None):
    """Check whether the given ``project`` has opened scenes

    :return Whether the user wants to restore scenes
    """
    if not data:
        return False

    ask = (f'The project currently has <b>{len(data)}</b> active scene(s). '
           'Do you want to open the scenes after project has been reloaded?')
    options = (QMessageBox.Yes | QMessageBox.No)
    reply = QMessageBox.question(parent, 'Scenes reload', ask,
                                 options, QMessageBox.No)
    if reply == QMessageBox.No:
        return False

    if reply == QMessageBox.Yes:
        return True


def show_trash_project_message(is_trashed, simple_name='', parent=None):
    """Check whether the given ``is_trashed`` should be toggled and show a
    messagebox to allow the user to confirm moving to trash or restoring from
    trash or cancel

    :return Whether the user wants to change the ``is_trashed`` attribute
    """
    if is_trashed:
        title = 'Restore from trash'
        text = ('Do you really want to <b>restore</b> this project <br><b>{}'
                '</b> from trash?').format(simple_name)
    else:
        title = 'Move to trash'
        text = ('Do you really want to <b>move</b> this project <br><b>{}</b>'
                ' to trash?').format(simple_name)

    result = QMessageBox.question(parent, title, text,
                                  QMessageBox.Yes | QMessageBox.No)
    if result == QMessageBox.Yes:
        return True
    return False


def show_no_configuration():
    """Broadcast event to show no configuration in configuration panel
    """
    broadcast_event(KaraboEvent.ShowConfiguration, {'proxy': None})


def run_macro(macro_model, parent=None):
    """
    Request to run a macro by validating and sending an instantiate.

    :param macro_model:  Model of the Macro.
    :type macro_model: karabo.common.project.macro.MacroModel
    :param parent: Parent for the QMessageBox.
    :type parent: QObject
    """
    macro_servers = get_macro_servers()
    if not macro_servers:
        messagebox.show_error("No (stable) Macro server found in system "
                              "topology. Macro cannot be started.")
        return

    if macro_model.modified:
        options = (QMessageBox.Yes | QMessageBox.No)
        title = "Run Unsaved Macro?"
        text = (
            "The Macro is not saved in the Project. Do you really want to "
            "run the Macro?")
        reply = QMessageBox.question(parent, title, text, options,
                                     QMessageBox.Yes)
        if reply == QMessageBox.No:
            return

    serverId = macro_servers[0]
    _run_macro(macro_model, serverId, parent)


def restart_macro(instance_id, handler, parent=None):
    """Restart macro after confirmation that macro should be restarted.

    Function will do nothing if user declined marco restart
    NOTE: macro should be instantiated"""
    ask = (f"The macro <b>{instance_id}</b> is already online. "
           "Do you want to restart the macro?")
    options = QMessageBox.Yes | QMessageBox.No
    reply = QMessageBox.question(
        parent, "Macro is already online", ask, options, QMessageBox.No)
    if reply != QMessageBox.Yes:
        return
    proxy = get_topology().get_device(instance_id, request=False)
    onShutdown(proxy, handler)


def run_macro_debug(macro_model, serverId, parent=None):
    """Request to run a macro by validating and sending an instantiate.

    This method doesn't ask for modification of the macro

    :param macro_model: Model of the Macro.
    :type macro_model: karabo.common.project.macro.MacroModel
    :param serverId: serverId of a development macro server.
    :type serverId: Str
    :param parent: Parent for the QMessageBox.
    :type parent: QObject
    """
    _run_macro(macro_model, serverId, parent)


def _run_macro(macro_model, serverId, parent):
    """Run a macro implementation to instantiate a macro"""
    report = validate_macro(macro_model.code)
    if report:
        fails = create_list_string(report)
        html = ("Please check the macro validation report and rename "
                f"function names or remove forbidden imports: {fails}")
        messagebox.show_error(html, title="The Macro cannot be started",
                              parent=parent)
        return

    instance_id = macro_model.instance_id
    h = Hash("code", macro_model.code,
             "module", macro_model.simple_name,
             "uuid", macro_model.uuid)

    # Backward compatibility, only servers with `MetaMacro` class of 2.16.X
    # have the project property
    attrs = get_topology().get_attributes(f"server.{serverId}")
    if attrs is None:
        messagebox.show_error(
            f"No Macro server found with serverId '{serverId}' in system "
            f"topology. Macro cannot be started.", parent=parent)
        return
    version = attrs.get("karaboVersion", "0.0.0")
    if version_compatible(version, 2, 16):
        h["project"] = macro_model.project_name
    get_network().onInitDevice(serverId, "MetaMacro", instance_id, h)


def _is_karabo3() -> bool:
    path = f"device.{KARABO_PROJECT_MANAGER}"
    attributes = get_topology().get_attributes(path)
    is_karabo3 = False
    if attributes is not None:
        version = attributes.get("karaboVersion", "0.0.0")
        is_karabo3 = version_compatible(version, 3, 0)
    return is_karabo3



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import mkColor
from pyqtgraph.exporters import ImageExporter as PgImageExporter

from karabo.native import Timestamp
from karabogui import util
from karabogui.singletons.api import get_config


class PlotDataExporter:
    def __init__(self, data_items):
        if len(data_items) == 1:
            x_data, y_data = data_items[0].getData()
            self._exporter = ArrayExporter(self._stacked_xy(x_data, y_data))
        else:
            saved = {curve.name(): self._stacked_xy(*curve.getData())
                     for curve in data_items}
            self._exporter = MultiArrayExporter(saved)

    def export(self):
        self._exporter.export()

    def _stacked_xy(self, x_data, y_data):
        if x_data is None and y_data is None:
            x_data, y_data = [], []

        # Adjust size if not same, usually happens with histogram plots,
        # wherein the last element of the x-axis can be removed
        x_size, y_size = len(x_data), len(y_data)
        if x_size != y_size:
            if x_size > y_size:
                x_data = x_data[:y_size]
            else:
                y_data = y_data[:x_size]

        return np.vstack((x_data, y_data))


class ArrayExporter:
    """Exporter for single numpy arrays. This is typically used for images and
       plots with single curves.

       For the case of the latter, the x- and y-components are aggregated into
       one array:
          x = np.array([1,2,3])
          y = np.array([4,5,6])
          data = [x, y]

       Data is saved as:
          np.save("foo.npy", data)

       And can be then accessed as:
          saved_data = np.load("foo.npy")
          saved_data
             >> np.array([[1,2,3],[4,5,6]])"""

    def __init__(self, data):
        if not isinstance(data, np.ndarray):
            raise TypeError(f"Data must be a Numpy array, not {type(data)}.")
        self._data = data

    def export(self):
        name = f"{Timestamp().toLocal()}_data.npy"
        filename = util.getSaveFileName(caption="Export Data",
                                        filter="Numpy File (*.npy)",
                                        suffix="npy",
                                        directory=get_config()["data_dir"],
                                        selectFile=name)

        if not filename:
            return

        if not filename.endswith(".npy"):
            filename = f"{filename}.npy"

        np.save(filename, self._data)


class MultiArrayExporter:
    """Exporter for multiple numpy arrays. This is typically used for
       multicurve plots. This utilizes proxy names as keys:

       data_1 = np.array([[1,2,3],[4,5,6]])
       data_2 = np.array([[7,8],[9,10]])


       Data is saved as:
          np.savez("foo.py", proxy_1=data_1, proxy_2=data_2)

       And can be then accessed as:
          saved_data = np.load("foo.npz")

          saved_data["proxy_1"]
            >> np.array([[1,2,3],[4,5,6]])

          saved_data["proxy_2"]
            >> np.array([[7,8],[9,10]])

       The x- and y-components are aggregated into one array, similar with
       `ArrayExporter`.
       """

    def __init__(self, zipped):
        for data in zipped.values():
            if not isinstance(data, np.ndarray):
                raise TypeError(f"Data must be a Numpy array, "
                                f"not {type(data)}.")
        self._zipped = zipped

    def export(self):
        name = f"{Timestamp().toLocal()}_data.npz"
        filename = util.getSaveFileName(caption="Export Data",
                                        filter="Numpy Zipped File (*.npz)",
                                        suffix="npz",
                                        directory=get_config()["data_dir"],
                                        selectFile=name)

        if not filename:
            return

        if not filename.endswith(".npz"):
            filename = f"{filename}.npz"

        np.savez(filename, **self._zipped)


class ImageExporter:
    def __init__(self, item):
        self._exporter = PgImageExporter(item)
        params = self._exporter.parameters()
        params["antialias"] = False
        params["background"] = mkColor("w")

    def export(self):
        """Exports the plot to an image"""
        # First we export the contents to a QImage
        qimage = self._exporter.export(toBytes=True)
        name = f"{Timestamp().toLocal()}_image.png"
        filename = util.getSaveFileName(caption="Save Snapshot",
                                        filter="PNG (*.png)",
                                        suffix="png",
                                        directory=get_config()["data_dir"],
                                        selectFile=name)

        if not filename:
            return

        if not filename.endswith(".png"):
            filename = f"{filename}.png"

        qimage.save(filename)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import io
import os
import re
from collections import defaultdict
from contextlib import redirect_stdout
from tempfile import mkstemp

from pycodestyle import Checker, StyleGuide
from pyflakes.api import check
from qtpy.Qsci import QsciAPIs, QsciScintilla
from qtpy.QtCore import (
    Property as pyqtProperty, QRegularExpression, Qt, Signal, Slot)
from qtpy.QtGui import QColor, QKeySequence
from qtpy.QtWidgets import QShortcut, QVBoxLayout, QWidget

from karabogui.widgets.find_toolbar import FindToolBar
from karabogui.widgets.scintilla_api import get_symbols
from karabogui.widgets.scintilla_lexer import PythonLexer

AUTO_COMPLETION_THRESHOLD = 3
LINE_LENGTH = 79
LINE_HIGHLIGHT = QColor("#FFFFCC")
MARGIN_BACKGROUND = QColor("#E0E0E0")
MARGIN_FOREGROUND = QColor("#A0A0A0")

# These are arbitrary numbers to represent an indicator style. There may be
# up to 32 types of indicator defined at a time. The first 8 are normally
# used by lexers.
ERROR_INDICATOR = 19
HIGHLIGHT_INDICATOR = 20
STYLE_ISSUE_INDICATOR = 21

# To match result from pycodestyle as filename:line_no:col_no:pep8Code message
# For example, "test.py:3:1: E302 expected 2 blank lines, found 1"
STYLE_REGEX = re.compile(r".*:(\d+):(\d+):\s+(.*)")

# To match result from pyflakes as filename:line_no:col_no: message
# For example, "test.py:1:1 'karabo.middlelayer.Int32' imported but unused"
FLAKE_REGEX = re.compile(r".*:(\d+):(\d+):?\s+(.*)")

IGNORED_STYLE_CODES = (
    "W292",  # No new line at the end
    "W293",  # Blank line contains whitespace
    )


class CodeBook(QWidget):
    """Python Code editor class based on Scintilla"""

    codeChanged = Signal()
    qualityChecked = Signal(bool)

    def __init__(self, code=None, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        code_editor = CodeEditor(parent=self)
        if code is not None:
            code_editor.setText(code)

        code_editor.textChanged.connect(self.codeChanged)
        code_editor.textChanged.connect(self._mayHighlight)
        code_editor.resultFound.connect(self.updateResultText)
        code_editor.qualityChecked.connect(self.qualityChecked)

        find_toolbar = FindToolBar(parent=self)
        find_toolbar.findRequested.connect(self._findNext)
        find_toolbar.highlightRequested.connect(self._highlight)
        find_toolbar.aboutToClose.connect(self._clearHighlight)
        find_toolbar.replaceRequested.connect(self._replace)
        find_toolbar.setVisible(False)

        find = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_F), self)
        find.activated.connect(self.showFindToolbar)

        replace = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_R), self)
        replace.activated.connect(self.showReplaceToolbar)

        self.find_toolbar = find_toolbar
        self.code_editor = code_editor
        layout.addWidget(self.find_toolbar)
        layout.addWidget(self.code_editor)

    @Slot()
    def showFindToolbar(self):
        self.find_toolbar.setVisible(True)
        selected_text = self.code_editor.selectedText()
        # Consider the text from previous search.
        search_text = self.find_toolbar.find_line_edit.text()
        match_case = self.find_toolbar.match_case.isChecked()
        if selected_text:
            self.find_toolbar.find_line_edit.setText(selected_text)
            search_text = selected_text
        self.code_editor.highlight(search_text, match_case)
        self.find_toolbar.find_line_edit.setFocus()
        self.find_toolbar.show_replace_tool_button.setChecked(False)
        self.find_toolbar.set_replace_widgets_visibility(False)

    @Slot()
    def showReplaceToolbar(self):
        self.showFindToolbar()
        self.find_toolbar.show_replace_tool_button.setChecked(True)

    @Slot(str, bool, bool)
    def _findNext(self, text, match_case, find_backward):
        self.code_editor.find_match(text, match_case, find_backward)

    @Slot(str, str, bool, bool)
    def _replace(self, text, new_text, match_case, replace_all):
        if replace_all:
            self.code_editor.replace_all(text, new_text, match_case)
        else:
            self.code_editor.replace_text(text, new_text, match_case)

    @Slot(int)
    def updateResultText(self, count):
        self.find_toolbar.setResultText(count)

    @Slot(str, bool)
    def _highlight(self, text, match_case):
        self.code_editor.highlight(text, match_case)

    @Slot()
    def _clearHighlight(self):
        self.code_editor.clearHighlight()

    @Slot()
    def _mayHighlight(self):
        """
        Update the search highlight on code change, if Find Toolbar is open.
        """
        if not self.find_toolbar.isVisible():
            return
        text = self.find_toolbar.find_line_edit.text()
        match_case = self.find_toolbar.match_case.isChecked()
        self._highlight(text, match_case)

    def increaseFontSize(self):
        self.code_editor.zoomIn()

    def decreaseFontSize(self):
        self.code_editor.zoomOut()

    def setReadOnly(self, value):
        """ Set the editor read only or not. """
        self.code_editor.setReadOnly(value)

    def moveCursorToLine(self, line_number, offset):
        self.code_editor.setCursorPosition(line_number, offset)

    def getEditorCode(self):
        """ Get the current code from the editor """
        return self.code_editor.text()

    def checkCode(self):
        self.code_editor.checkCodeQuality()

    def clearIndicators(self):
        self.code_editor.clearAnnotations()
        self.code_editor.clearAllIndicators()

    def clear(self):
        self.code_editor.clear()


class CodeEditor(QsciScintilla):
    """
    Python Code editor class.
    """

    resultFound = Signal(int)
    qualityChecked = Signal(bool)

    def __init__(self, parent=None, use_api=True):
        super().__init__(parent)

        # Indentation
        self.setTabWidth(4)
        self.setAutoIndent(True)
        self.setIndentationsUseTabs(False)
        self.setIndentationGuides(True)
        self.setBackspaceUnindents(True)
        self.setIndentationGuidesBackgroundColor(MARGIN_BACKGROUND)
        self.setIndentationGuidesForegroundColor(MARGIN_BACKGROUND)

        # Auto completion
        self.setAutoCompletionSource(QsciScintilla.AcsAll)
        self.setAutoCompletionReplaceWord(True)
        self.setAutoCompletionCaseSensitivity(False)
        self.autoCompleteFromDocument()
        self.setAutoCompletionThreshold(AUTO_COMPLETION_THRESHOLD)

        # Highlight the current line
        self.setCaretLineVisible(True)
        self.setCaretLineBackgroundColor(LINE_HIGHLIGHT)

        # lexer
        lexer = PythonLexer(parent=self)

        # Show line numbers
        font = lexer.font(PythonLexer.Default)
        self.setMarginsFont(font)
        self.setMarginWidth(0, "0000")
        self.setMarginLineNumbers(0, True)
        self.setMarginsBackgroundColor(MARGIN_BACKGROUND)
        self.setMarginsForegroundColor(MARGIN_FOREGROUND)

        # Brace matching: highlight the brace when the cursor is on left or
        # on right
        self.setBraceMatching(QsciScintilla.SloppyBraceMatch)

        # Code folding
        self.setFolding(QsciScintilla.PlainFoldStyle)
        self.setFoldMarginColors(MARGIN_BACKGROUND, MARGIN_BACKGROUND)

        # A vertical line to indicate the line-length limit
        self.setEdgeMode(QsciScintilla.EdgeLine)
        self.setEdgeColumn(LINE_LENGTH)
        self.setEdgeColor(QColor("gray"))

        if use_api:
            self._apis = QsciAPIs(lexer)
            for symbol in get_symbols():
                self._apis.add(symbol)
            self._apis.prepare()

        self.setLexer(lexer)

        self._highlights = []
        self.indicatorDefine(self.StraightBoxIndicator, HIGHLIGHT_INDICATOR)
        self.indicatorDefine(self.SquiggleLowIndicator, ERROR_INDICATOR)
        self.indicatorDefine(self.SquiggleLowIndicator, STYLE_ISSUE_INDICATOR)
        self.setIndicatorForegroundColor(QColor("red"), ERROR_INDICATOR)
        self.setIndicatorForegroundColor(QColor("blue"), STYLE_ISSUE_INDICATOR)
        self.has_annotation = False
        self.hit_count = 0

    @Slot(str, bool, bool)
    def find_match(self, text, match_case, find_backward):
        """
        Select the first match from the current cursor position.
        Return True if found else False.
        """
        regular_expression = False
        word_only = False
        wrap = True
        find_forward = not find_backward
        line, index = self.getCursorPosition()
        if find_backward:
            # Move cursor behind by one, so that backward search doesn't get
            # stuck on the just single hit.
            index -= 1
        found = self.findFirst(text, regular_expression, match_case, word_only,
                               wrap, find_forward, line, index)
        if not found:
            self.clearHighlight()
        return found

    def replace_text(self, text, new_text, match_case):
        """
        Replace the first hit of search text with 'new_text' and select the
        next hit. If the search text is not selected already, just select
        the next hit.
        """
        if self.hasSelectedText():
            selected_text = self.selectedText()
            if not match_case:
                selected_text = selected_text.lower()
                text = text.lower()
            if selected_text == text:
                self.replaceSelectedText(new_text)
        self.find_match(text, match_case, find_backward=False)

    def replace_all(self, text, new_text, match_case):
        """
        Recursively replace a text with a new text. Replace all would work
        only if the search text is already highlighted - which always
        happen first in the workflow.
        """
        found = self.find_match(text, match_case, find_backward=False)
        if found:
            self.beginUndoAction()
            for _ in range(self.hit_count):
                if self.findNext():
                    self.replace(new_text)
            self.endUndoAction()

    @Slot(str, bool)
    def highlight(self, text, match_case):
        """ Highlight all the occurrence of the given text."""
        self.clearHighlight()
        if text == "":
            return
        options = QRegularExpression().patternOptions()
        if not match_case:
            options = options | QRegularExpression.CaseInsensitiveOption
        pattern = QRegularExpression(text, options)
        if not pattern.isValid():
            return
        matches = pattern.globalMatch(self.text())

        hit_count = 0
        while matches.hasNext():
            match = matches.next()
            if match.capturedLength() != len(pattern.pattern()):
                # Avoid matching regular expression.
                continue
            start = match.capturedStart()
            end = match.capturedEnd()
            _range = self._range_from_position(start, end)
            line_start, index_start, line_end, index_end = _range

            selection = {"line_start": line_start,
                         "index_start": index_start,
                         "line_end": line_end,
                         "index_end": index_end}
            self._highlights.append(selection)
            self.fillIndicatorRange(
                line_start, index_start, line_end, index_end,
                HIGHLIGHT_INDICATOR)
            hit_count += 1
        self.hit_count = hit_count
        self.resultFound.emit(hit_count)

    def clearHighlight(self):
        for selection in self._highlights:
            self.clearIndicatorRange(
                selection["line_start"],
                selection["index_start"],
                selection["line_end"],
                selection["index_end"],
                HIGHLIGHT_INDICATOR
            )
        self._highlights.clear()
        self.hit_count = 0
        self.resultFound.emit(self.hit_count)

    def _range_from_position(self, start_position, end_position):
        """
        Convert a position which is globally indexed from the start of the
        text to a combination of line number and index from the start of the
        line.
        """
        start_line, start_offset = self.lineIndexFromPosition(start_position)
        end_line, end_offset = self.lineIndexFromPosition(end_position)
        return start_line, start_offset, end_line, end_offset

    def checkCodeQuality(self):
        """ Run 'pyflakes' and 'pycodestyle' to check the code quality"""
        self.clearAnnotations()
        self.clearAllIndicators()

        code = self.text()
        indicators = defaultdict(list)

        flake_feedback = check_flake(code)
        for line_number, feedback in flake_feedback.items():
            indicators[line_number].extend(feedback)

        style_feedback = check_style(code)
        for line_number, feedback in style_feedback.items():
            indicators[line_number].extend(feedback)
        error = False
        if flake_feedback or style_feedback:
            self._annotate_code(indicators)
            error = True
            self.has_annotation = True
        self.qualityChecked.emit(error)

    def _annotate_code(self, indicators):
        for line_number, descriptions in indicators.items():
            messages = []
            # Editor follows zero based indexing.
            line_number = line_number - 1
            for description in descriptions:
                message = description.get("message")
                messages.append(message)
                line_text = self.text(line_number)
                col_end = len(line_text)
                col_start = col_end - len(line_text.strip()) - 1
                indicator = (ERROR_INDICATOR if description.get(
                    "is_error", False) else STYLE_ISSUE_INDICATOR)
                self.fillIndicatorRange(line_number, col_start,
                                        line_number, col_end, indicator)
            message = "\n".join(messages)
            self.annotate(line_number,  message, self.annotationDisplay())

    def clearAllIndicators(self):
        line_start = 0
        col_start = 0
        line_end = self.lines()
        col_end = len(self.text(line_end-1))
        self.clearIndicatorRange(
            line_start, col_start, line_end, col_end, ERROR_INDICATOR
        )
        self.clearIndicatorRange(
            line_start, col_start, line_end, col_end, STYLE_ISSUE_INDICATOR
        )
        self.has_annotation = False

    @pyqtProperty(str)
    def textForSquish(self):
        """
        This is only for the Squish test. Squish can't normally access a
        method in a custom Qt Widget, unless it is a Slot or Property.
        Wrapping this method with pyqtProperty so that Squish can read
        access the editor text as a property.
        """
        return super().text()


class FlakeReporter:
    """
    This class imitates 'pyflakes.reporter.Reporter' class but
    additionally stores the report as 'log' instead of just printing out.
    """
    def __init__(self):
        self.log = defaultdict(list)

    def unexpectedError(self, filename, message):
        self.log[0].append({"filename": filename,
                            "message": str(message)
                            })

    def syntaxError(self, filename, message, line_no, column, source):
        self.log[line_no].append({
            "message": str(message),
            "is_error": True,
        })

    def flake(self, message):
        matcher = FLAKE_REGEX.match(str(message))
        if matcher:
            line_no, col, msg = matcher.groups()
            line_no = int(line_no)
            self.log[line_no].append({
                "message": msg,
            })


def check_flake(code):
    reporter = FlakeReporter()
    # Skipping the filename as it is only used to mention in error report.
    check(code, filename="", reporter=reporter)
    return reporter.log


def check_style(code):
    code_fd, code_filename = mkstemp()
    os.close(code_fd)

    with open(code_filename, "w") as fh:
        fh.write(code)

    style = StyleGuide(config_file="")  # Optionally can add config file here.
    ignore_codes = style.options.ignore + IGNORED_STYLE_CODES
    style.options.ignore = ignore_codes
    checker = Checker(code_filename, options=style.options)
    with redirect_stdout(io.StringIO()) as f:
        checker.check_all()
    results = f.getvalue()

    os.remove(code_filename)

    style_feedback = defaultdict(list)
    for result in results.split("\n"):
        matcher = STYLE_REGEX.match(result)
        if matcher:
            line_no, col, msg = matcher.groups()
            line_no = int(line_no)
            code, description = msg.split(" ", 1)
            style_feedback[line_no].append({
                "message": description,
                "code": code,
            })
    return style_feedback



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialog

from .utils import get_dialog_ui

DEFAULT_FORMAT = "f"
DEFAULT_FRACTIONS = "3"


class FormatFmtDialog(QDialog):

    def __init__(self, value=None, fmt=DEFAULT_FORMAT,
                 decimals=DEFAULT_FRACTIONS, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui("format_fmt.ui"), self)
        self.value = value
        index = self.combo_fmt.findText(fmt)
        self.combo_fmt.setCurrentIndex(index)

        index = self.combo_decimals.findText(decimals)
        self.combo_decimals.setCurrentIndex(index)

        # Add triggers to change preview
        self.combo_fmt.currentIndexChanged.connect(self._preview)
        self.combo_decimals.currentIndexChanged.connect(self._preview)
        # Update preview to reflect current format
        self._preview()

    @Slot()
    def _preview(self):
        """Preview the value value formatting"""
        combo_fmt = self.combo_fmt.currentText()
        combo_decimals = self.combo_decimals.currentText()

        preview = f"{{:.{combo_decimals}{combo_fmt}}}"
        if self.value is not None:
            try:
                preview = preview.format(self.value)
            except Exception:
                preview = "Not a valid formatting .."
        self.preview_label.setText(preview)

    @property
    def fmt(self):
        return self.combo_fmt.currentText()

    @property
    def decimals(self):
        return self.combo_decimals.currentText()



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 26, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from functools import partial

from qtpy.QtCore import QSize
from qtpy.QtWidgets import QDialog, QVBoxLayout, QWidget

from karabo.common.project.api import ProjectModel
from karabogui import icons
from karabogui.access import AccessRole, access_role_allowed
from karabogui.actions import KaraboAction, build_qaction
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts)
from karabogui.logger import get_logger
from karabogui.project.dialog.project_handle import NewProjectDialog
from karabogui.project.restore import add_restore, get_restore_data
from karabogui.project.utils import (
    load_project, load_project_with_item, maybe_save_modified_project,
    reload_project, save_object, show_modified_project_message,
    show_reload_scene_project_message)
from karabogui.project.view import ProjectView
from karabogui.singletons.api import get_db_conn
from karabogui.util import get_spin_widget
from karabogui.widgets.toolbar import ToolBar

from .base import BasePanelWidget
from .searchwidget import SearchBar


class ProjectPanel(BasePanelWidget):
    """ A dockable panel which contains a view of the project
    """

    def __init__(self):
        super().__init__("Projects")
        # Register for broadcast events.
        # This object lives as long as the app. No need to unregister.
        event_map = {
            KaraboEvent.NetworkConnectStatus: self._event_network,
            KaraboEvent.DatabaseIsBusy: self._event_db_busy,
            KaraboEvent.ProjectFilterUpdated: self._event_filter_updated,
            KaraboEvent.AccessLevelChanged: self._event_access_level,
            KaraboEvent.LoginUserChanged: self._event_access_level,
            KaraboEvent.ShowProjectDevice: self._event_show_project_device,
            KaraboEvent.ShowProjectModel: self._event_select_project_model,
        }
        register_for_broadcasts(event_map)

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        widget = QWidget(self)
        main_layout = QVBoxLayout(widget)
        main_layout.setContentsMargins(2, 2, 2, 2)

        self.tree_view = ProjectView(parent=widget)
        self.sbar = SearchBar(parent=widget)
        model = self.tree_view.model()
        self.sbar.setModel(model)

        main_layout.addWidget(self.sbar)
        main_layout.addWidget(self.tree_view)

        return widget

    def __repr__(self):
        qt_model = self.tree_view.model()
        root = qt_model.root_model
        project_name = root.simple_name if root is not None else None
        return f"<ProjectPanel project={project_name}>"

    def toolbars(self):
        """This should create and return one or more `ToolBar` instances needed
        by this panel.
        """
        self._toolbar_actions = []
        new = KaraboAction(
            icon=icons.new, text="&New Project",
            tooltip="Create a New (Sub)project",
            triggered=_project_new_handler,
            name="new"
        )
        load = KaraboAction(
            icon=icons.load, text="&Load Project",
            tooltip="Load an Existing Project",
            triggered=_project_load_handler,
            name="load"
        )
        load_with_item = KaraboAction(
            icon=icons.filein, text="Find and &Load Project with Item Names",
            tooltip="Find and Load Project with Item Names",
            triggered=_project_with_device_load_handler,
            name="load_with_item"
        )
        save = KaraboAction(
            icon=icons.save, text="&Save Project",
            tooltip="Save Project Snapshot",
            triggered=_project_save_handler,
            name="save"
        )
        reload = KaraboAction(
            icon=icons.reset, text="&Reload Project",
            tooltip="Reload the current project",
            triggered=_project_reload_handler,
            name="reload"
        )
        trash = KaraboAction(
            icon=icons.delete, text="&Declare as trashed or untrashed",
            tooltip=("Mark a project as trashed or untrashed depending on "
                     "the project state."),
            triggered=_project_trash_handler,
            name="declare"
        )

        for k_action in (new, load, load_with_item, save, reload, trash):
            q_ac = build_qaction(k_action, self)
            q_ac.setEnabled(True)
            q_ac.setVisible(False)
            tree_view = self.tree_view
            q_ac.triggered.connect(partial(k_action.triggered,
                                           tree_view))
            self._toolbar_actions.append(q_ac)

        toolbar = ToolBar(parent=self)
        for ac in self._toolbar_actions:
            toolbar.addAction(ac)

        # Add a spinner which is visible whenever database is processing
        toolbar.add_expander()
        spin_widget = get_spin_widget(icon='wait-black',
                                      scaled_size=QSize(20, 20),
                                      parent=toolbar)
        spin_widget.setVisible(False)
        self.spin_action = toolbar.addWidget(spin_widget)

        return [toolbar]

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_access_level(self, data):
        self._access_toolbar()

    def _event_db_busy(self, data):
        loading_failed = data.get('loading_failed', False)
        is_processing = data['is_processing']
        if loading_failed:
            self._visible_partial_toolbar()
        else:
            self._set_toolbar_visible(not is_processing)
            self.sbar.reset(not is_processing)
        self.spin_action.setVisible(is_processing)

    def _event_filter_updated(self, data):
        self.sbar.reset(data['status'])

    def _event_network(self, data):
        status = data['status']
        if not status:
            # Don't show projects when there's no server connection
            self.tree_view.destroy_model()

        self._set_toolbar_visible(status)
        if not status:
            self.sbar.reset(status)

    def _event_show_project_device(self, data):
        """Event to show a project device """
        model = self.tree_view.model()
        nodes = model.findNodes(data.get("deviceId"), full_match=True)
        assert len(nodes) <= 1
        if nodes:
            # Select first entry
            model.selectNode(nodes[0])

    def _event_select_project_model(self, data):
        """Event to show a project device """
        model = self.tree_view.model()
        model.selectModel(data.get("model"))

    # -----------------------------------------------------------------------

    def _set_toolbar_visible(self, enable):
        for qaction in self._toolbar_actions:
            qaction.setVisible(enable)

    def _visible_partial_toolbar(self):
        """ Project loading failed, restrict options
        """
        for qaction in self._toolbar_actions:
            if qaction.objectName() in ("new", "load"):
                qaction.setVisible(True)

    def _access_toolbar(self):
        enable = access_role_allowed(AccessRole.PROJECT_EDIT)
        for qaction in self._toolbar_actions:
            if qaction.objectName() in ("new", "save", "declare"):
                qaction.setEnabled(enable)


# ------------------------------------------------------------------------
# Helper functions


def _project_load_handler(project_view):
    """ Load a project model (`ProjectViewItemModel`) and assign it to the
    `ProjectViewItemModel` of the given `project_view`

    :param project_view: The `ProjectView` of the panel
    """
    item_model = project_view.model()
    # Check for modififications before showing dialog
    root_model = item_model.root_model
    if not maybe_save_modified_project(root_model, parent=project_view):
        return

    project = load_project(parent=project_view)
    if project is not None:
        item_model.root_model = project


def _project_with_device_load_handler(project_view):
    """ Load a project with a user specified device, assigning it to the
    `ProjectViewItemModel` of the given `project_view`.

    :param project_view: The `ProjectView` of the panel
    """
    item_model = project_view.model()
    # Check for modififications before showing dialog
    root_model = item_model.root_model
    if not maybe_save_modified_project(root_model, parent=project_view):
        return

    project = load_project_with_item(parent=project_view)
    if project is not None:
        item_model.root_model = project


def _project_reload_handler(project_view):
    """ Reload a project model (`ProjectViewItemModel`)

    :param project_view: The `ProjectView` of the panel
    """
    item_model = project_view.model()
    root_model = item_model.root_model
    if root_model is None:
        return
    if not show_modified_project_message(root_model, parent=project_view):
        return

    data = get_restore_data()
    restore = show_reload_scene_project_message(data)
    project = reload_project(root_model)
    # We created a new project object!
    if project is not None:
        item_model.root_model = project
        text = (f"Reloading project <b>{root_model.simple_name}</b> from "
                "project database")
        get_logger().info(text)
        if restore:
            add_restore(project=project, scene_data=data)


def _project_new_handler(project_view):
    """ Create a new project model (`ProjectViewItemModel`) and assign it to
    the given `project_view`

    :param project_view: The `ProjectView` of the panel
    """
    item_model = project_view.model()
    # Check for modififications before showing dialog
    root_model = item_model.root_model
    if not maybe_save_modified_project(root_model, parent=project_view):
        return

    dialog = NewProjectDialog(parent=project_view)
    if dialog.exec() == QDialog.Accepted:
        # Set domain
        get_db_conn().default_domain = dialog.domain
        # This overwrites the current model
        model = ProjectModel(simple_name=dialog.simple_name, initialized=True,
                             modified=True)
        item_model.root_model = model
        text = f"Creating new project <b>{dialog.simple_name}</b>"
        get_logger().info(text)
        broadcast_event(KaraboEvent.ProjectName,
                        {"simple_name": model.simple_name})


def _project_save_handler(project_view):
    """ Save the project model (`ProjectViewItemModel`) of the given
    `project_view`

    :param project_view: The `ProjectView` of the panel
    """
    item_model = project_view.model()
    root_model = item_model.root_model
    if root_model is not None:
        save_object(root_model)
        text = (f"Request to save project <b>{root_model.simple_name}</b> in "
                "the project database")
        get_logger().info(text)
        broadcast_event(KaraboEvent.ProjectName,
                        {"simple_name": root_model.simple_name})


def _project_trash_handler(project_view):
    """ Move the project model (`ProjectViewItemModel`) of the given
    `project_view` to trash

    :param project_view: The `ProjectView` of the panel
    """
    item_model = project_view.model()
    root_model = item_model.root_model
    if root_model is not None:
        project_view.update_is_trashed(project=root_model,
                                       project_controller=None)
        text = ("Changing the <b>trash</b> attribute of the project "
                f"<b>{root_model.simple_name}</b>")
        get_logger().warning(text)



import pytest
from qtpy.QtCore import QPoint, Qt
from qtpy.QtGui import QColor, QFont
from qtpy.QtTest import QTest
from qtpy.QtWidgets import (
    QDialog, QGraphicsLineItem, QGraphicsPixmapItem, QGraphicsRectItem,
    QGraphicsSimpleTextItem)

from karabo.native import Hash, HashList
from karabogui.binding.api import DeviceClassProxy, build_binding
from karabogui.dialogs.logbook_drawing_tools import (
    CropTool, LineTool, RectTool, TextTool)
from karabogui.dialogs.logbook_preview import LogBookPreview
from karabogui.panels.api import ConfigurationPanel
from karabogui.singletons.api import get_config
from karabogui.singletons.configuration import Configuration
from karabogui.singletons.mediator import Mediator
from karabogui.singletons.network import Network
from karabogui.testing import get_simple_schema, singletons

TITLE = ("{}: <ConfigurationPanel proxy=<DeviceClassProxy classId=Simple "
         "serverId=test_server>>")


@pytest.fixture(scope="function")
def dialog(gui_app):
    mediator = Mediator()
    config = Configuration()
    with singletons(mediator=mediator, configuration=config):
        panel = ConfigurationPanel()

        schema = get_simple_schema()
        binding = build_binding(schema)
        proxy = DeviceClassProxy(server_id="test_server", binding=binding)
        panel._show_configuration(proxy)
        dialog = LogBookPreview(parent=panel)
        yield dialog
        dialog._event_network({"status": False})


def test_logbook_preview(dialog):
    """Test the logbook preview with a configuration panel"""
    assert dialog.stackedWidget.currentIndex() == 0
    assert dialog.table.columnCount() == 2
    assert dialog.table.rowCount() == 3
    dialog.done(0)


def test_image_preview(dialog, mocker):
    assert dialog.combo_datatype.currentText() == "Image"
    network = Network()
    with singletons(network=network):
        saveLogBook = mocker.patch.object(network, "onSaveLogBook")
        dialog.combo_title_style.setCurrentIndex(1)
        dialog.done(1)
        assert saveLogBook.call_count == 1
        _, args = saveLogBook.call_args
        assert args["dataType"] == "text_image"
        assert args["title"] == TITLE.format("Image")
        assert args["data"].startswith("data:image/png;base64,")
        assert not args["caption"]


def test_logbook_table_preview(dialog, mocker):
    dialog.combo_datatype.setCurrentIndex(1)
    assert dialog.combo_datatype.currentText() == "Data"
    assert dialog.stackedWidget.currentIndex() == 1
    assert all([checkbox.isChecked() for checkbox in dialog.checkboxes])
    dialog._deselect_all()
    assert not any([checkbox.isChecked() for checkbox in dialog.checkboxes])
    dialog._select_all()
    assert all([checkbox.isChecked() for checkbox in dialog.checkboxes])

    dialog.combo_title_style.setCurrentIndex(1)

    network = Network()
    with singletons(network=network):
        saveLogBook = mocker.patch.object(network, "onSaveLogBook")
        dialog.done(1)
        assert saveLogBook.call_count == 1
        _, args = saveLogBook.call_args
        assert args["dataType"] == "hash"
        assert args["title"] == TITLE.format("Data")
        for item in ("foo", "bar", "charlie"):
            assert item in args["data"]

        # Unselected row should not be saved
        dialog.checkboxes[0].setChecked(False)
        dialog.done(1)
        _, args = saveLogBook.call_args
        data = args["data"]
        assert "foo" not in data
        for item in ("bar", "charlie"):
            assert item in data


def test_save_button(dialog):
    """Test the enabled state of the Save button"""
    assert dialog.ok_button.text() == "Save"
    assert dialog.combo_datatype.currentText() == "Image"
    # Disabled when no proposals.
    assert not dialog.ok_button.isEnabled()

    h_list = HashList()
    h_list.append(Hash("name", "test_stream",
                       "topics", ["Logbook", "Summary"]))
    dialog._event_destinations(h_list)
    assert dialog.ok_button.isEnabled()

    dialog.combo_datatype.setCurrentIndex(1)
    assert dialog.combo_datatype.currentText() == "Data"
    assert dialog.ok_button.isEnabled()

    # Disabled on deselecting all properties.
    dialog._deselect_all()
    assert not dialog.ok_button.isEnabled()

    # Enabled always for Image.
    dialog.combo_datatype.setCurrentIndex(0)
    assert dialog.combo_datatype.currentText() == "Image"
    assert dialog.ok_button.isEnabled()

    # Enabled when at least one property is selected in the table.
    dialog.combo_datatype.setCurrentIndex(1)
    assert not dialog.ok_button.isEnabled()
    dialog.checkboxes[0].setChecked(True)
    assert dialog.ok_button.isEnabled()

    # Disabled when no property is selected.
    dialog.checkboxes[0].setChecked(False)
    assert not dialog.ok_button.isEnabled()

    # Disabled when no topic is set
    destination_widget = dialog._destinations[0]
    destination_widget.combo_topic.setCurrentText("")
    assert not dialog.ok_button.isEnabled()


def test_title(dialog, mocker):
    default_title = ("<ConfigurationPanel proxy=<DeviceClassProxy "
                     "classId=Simple serverId=test_server>>")
    assert default_title == dialog.title_line_edit.placeholderText()
    network = Network()
    with singletons(network=network):
        saveLogBook = mocker.patch.object(network, "onSaveLogBook")
        dialog.combo_title_style.setCurrentIndex(1)
        dialog.done(1)
        _, args = saveLogBook.call_args
        assert args["title"] == TITLE.format("Image")
        assert args["headerType"] == "bold"

        custom_title = "Configuration Panel"
        dialog.title_line_edit.setText(custom_title)
        dialog.done(1)
        _, args = saveLogBook.call_args
        assert args["title"] == f"Image: {custom_title}"

        # Table
        dialog.combo_datatype.setCurrentIndex(1)
        dialog.combo_title_style.setCurrentIndex(2)
        dialog.done(1)
        _, args = saveLogBook.call_args
        assert args["title"] == f"Data: {custom_title}"
        assert args["headerType"] == "expandable"

        dialog.title_line_edit.clear()
        dialog.done(1)
        _, args = saveLogBook.call_args
        assert args["title"] == TITLE.format("Data")
        dialog.combo_title_style.setCurrentIndex(0)
        dialog.done(1)
        _, args = saveLogBook.call_args
        assert args["title"] == TITLE.format("Data")


def test_toolbar(dialog):
    toolbar = dialog.drawing_toolbar
    assert toolbar.isVisibleTo(dialog)
    actions = toolbar.actions()
    assert len(actions) == 7
    # First item is a label with  warning text
    expected = ["", "Change annotation color", "Draw Line", "Draw Rectangle",
                "Add Text to the Image", "Delete item", "Crop image"]
    assert [action.toolTip() for action in actions] == expected


def test_drawing_tools(dialog, mocker):
    canvas = dialog.canvas
    assert canvas.drawing_tool is None
    assert len(canvas.items()) == 1
    assert isinstance(canvas.items()[0], QGraphicsPixmapItem)

    can_draw = mocker.patch.object(dialog.canvas, "_can_draw")
    can_draw.return_value = True

    pos = QPoint(100, 100)
    widget = dialog.view.viewport()
    line_tool = LineTool()
    canvas.set_drawing_tool(line_tool)
    QTest.mousePress(widget, Qt.LeftButton, Qt.NoModifier)
    QTest.mouseMove(widget, pos)
    QTest.mouseRelease(widget, Qt.LeftButton, Qt.NoModifier)
    assert len(canvas.items()) == 2
    assert isinstance(canvas.items()[0], QGraphicsLineItem)
    assert isinstance(canvas.items()[1], QGraphicsPixmapItem)

    rect_tool = RectTool()
    canvas.set_drawing_tool(rect_tool)
    QTest.mousePress(widget, Qt.LeftButton, Qt.NoModifier, pos)
    QTest.mouseMove(widget, QPoint(200, 200))
    QTest.mouseRelease(widget, Qt.LeftButton, Qt.NoModifier)
    assert len(canvas.items()) == 3
    assert isinstance(canvas.items()[0], QGraphicsRectItem)
    assert isinstance(canvas.items()[1], QGraphicsLineItem)
    assert isinstance(canvas.items()[2], QGraphicsPixmapItem)

    text_tool = TextTool()
    text_dialog = mocker.patch(
        "karabogui.dialogs.logbook_drawing_tools.TextDialog")
    text_dialog().exec.return_value = QDialog.Accepted
    text_dialog().text_font = QFont()
    text_dialog().text_color = QColor("black")
    text_dialog().text = "Hello Karabo"
    canvas.set_drawing_tool(text_tool)
    QTest.mousePress(widget, Qt.LeftButton, Qt.NoModifier, pos)
    assert len(canvas.items()) == 3
    QTest.mouseRelease(widget, Qt.LeftButton, Qt.NoModifier, pos)
    assert len(canvas.items()) == 4
    assert isinstance(canvas.items()[0], QGraphicsSimpleTextItem)
    assert isinstance(canvas.items()[1], QGraphicsRectItem)
    assert isinstance(canvas.items()[2], QGraphicsLineItem)
    assert isinstance(canvas.items()[3], QGraphicsPixmapItem)
    assert canvas.items()[0].text() == "Hello Karabo"

    # Clear the active drawing tool.
    canvas.set_drawing_tool(None)
    assert canvas.drawing_tool is None
    can_draw.return_value = False
    QTest.mousePress(widget, Qt.LeftButton, Qt.NoModifier, pos)
    QTest.mouseMove(widget, QPoint(200, 200))
    QTest.mouseRelease(widget, Qt.LeftButton, Qt.NoModifier)
    assert len(canvas.items()) == 4

    tool = CropTool()
    canvas.set_drawing_tool(tool)
    assert canvas.drawing_tool is not None
    can_draw.return_value = False
    QTest.mousePress(widget, Qt.LeftButton, Qt.NoModifier, pos)
    assert canvas.drawing_tool is None


def test_topics(dialog):
    """Test the topics combobox is updated when the stream selection changes"""
    h_list = HashList()
    stream1 = "12345"
    topics1 = ["Logbook", "Setup", "Day Shift"]

    stream2 = "34567"
    topics2 = ["Data Analysis", "Detector status"]

    h_list.append(Hash("name", stream1, "topics", topics1))
    h_list.append(Hash("name", stream2, "topics", topics2))
    dialog._event_destinations(h_list)
    destination_widget = dialog._destinations[0]
    destination_widget.combo_stream.setCurrentText(stream2)
    combo = destination_widget.combo_topic
    assert [combo.itemText(i) for i in range(combo.count())] == topics2

    destination_widget.combo_stream.setCurrentText(stream1)
    assert [combo.itemText(i) for i in range(combo.count())] == topics1


def test_restore_stream_topic(dialog):
    h_list = HashList()
    stream1 = "12345"
    topics1 = ["Logbook", "Setup", "Day Shift"]

    stream2 = "34567"
    topics2 = ["Data Analysis", "Summary", "Detector status",
               "Sample status", "Beam Status"]

    stream3 = "98765"
    topics3 = ["Logbook", "RunSummary", "ShiftSummary"]

    h_list.append(Hash("name", stream1, "topics", topics1))
    h_list.append(Hash("name", stream2, "topics", topics2))
    h_list.append(Hash("name", stream3, "topics", topics3))

    default_stream = "34567"
    default_topic = "Summary"
    get_config()["logbook_stream"] = default_stream
    get_config()["logbook_topic"] = default_topic

    dialog._event_destinations(h_list)

    destination_widget = dialog._destinations[0]
    assert destination_widget.combo_stream.currentText() == default_stream
    assert destination_widget.combo_topic.currentText() == default_topic


def test_create_new_topic(dialog):
    """
    Test new topic creation. Verify that create_topic button
    enables/disables the edit mode on topic combobox.
    """
    destination_widget = dialog._destinations[0]
    destination_widget.create_topic.clicked.emit(True)
    assert destination_widget.combo_topic.isEditable()

    line_edit = destination_widget.combo_topic.lineEdit()
    line_edit.setText("NewTopic")
    destination_widget.combo_topic.lineEdit().editingFinished.emit()
    assert destination_widget.combo_topic.findText("NewTopic") != -1
    assert destination_widget.combo_topic.currentText() == "NewTopic"

    destination_widget.create_topic.setChecked(False)
    assert not destination_widget.combo_topic.isEditable()

    # Do not allow empty topic
    destination_widget.create_topic.clicked.emit(True)
    destination_widget.combo_topic.lineEdit().setText("")
    destination_widget.combo_topic.lineEdit().editingFinished.emit()
    assert destination_widget.combo_topic.currentText() == "NewTopic"


def test_concurrency(dialog):
    """
    The KaraboEvent.ActiveDestinations should not duplicate the streams
    in the combobox. This happens when multiple dialogs are opened.
    """
    h_list = HashList()
    h_list.append(Hash("name", "test_stream",
                       "topics", ["Logbook", "Summary"]))
    dialog._event_destinations(h_list)
    destination_widget = dialog._destinations[0]
    assert destination_widget.combo_stream.count() == 1

    dialog._event_destinations(h_list)
    assert destination_widget.combo_stream.count() == 1



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QModelIndex, QSize, Qt, Slot
from qtpy.QtGui import QIcon, QStandardItem, QStandardItemModel
from qtpy.QtWidgets import (
    QAction, QDialog, QHBoxLayout, QListView, QToolButton, QWidget)
from traits.api import Instance

import karabogui.icons as icons
from karabo.common.scenemodel.api import DisplayIconCommandModel
from karabogui.binding.api import SlotBinding
from karabogui.controllers.api import (
    BaseBindingController, is_proxy_allowed, register_binding_controller)

BUTTON_ICONS = {
    "Left": icons.arrowFancyLeft,
    "Right": icons.arrowFancyRight,
    "Up": icons.arrowFancyUp,
    "Down": icons.arrowFancyDown,
    "Start": icons.mediaStart,
    "Stop": icons.mediaStop,
    "Pause": icons.mediaPause,
    "Reset": icons.mediaBackward,
    "Acquire": icons.mediaRecord,
    "On": icons.on,
    "Off": icons.off,
    "Abort": icons.stop,
    "Halt": icons.halt
}

NO_SELECTION = icons.imageMissing
QPADDING = 2


class DynamicToolButton(QToolButton):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAutoRaise(True)
        self.setMinimumSize(20, 20)
        self.setEnabled(False)

    def resizeEvent(self, event):
        width, height = self.width(), self.height()
        self.setIconSize(QSize(width - QPADDING, height - QPADDING))
        super().resizeEvent(event)


class IconSelectionDialog(QDialog):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setModal(False)
        self.setWindowTitle("Select icon")
        self.setWindowFlags(Qt.Window |
                            Qt.WindowTitleHint |
                            Qt.CustomizeWindowHint)

        # Variables
        self.icon = NO_SELECTION
        self.name = ""

        # Setup list view
        self.list_view = QListView(self)
        self.list_view.setMinimumSize(400, 600)
        model = QStandardItemModel(self.list_view)

        # we add all items with a preview icon
        for key, icon in sorted(BUTTON_ICONS.items()):
            listItem = QStandardItem(key)
            listItem.setData(icon, Qt.DecorationRole)
            listItem.setData(icon, Qt.UserRole + 1)
            model.appendRow(listItem)
        self.list_view.setModel(model)
        self.list_view.doubleClicked.connect(self.handleDoubleClick)

    @Slot(QModelIndex)
    def handleDoubleClick(self, index):
        icon = index.data(Qt.UserRole + 1)
        if icon is not None:
            self.icon = icon
            self.name = index.data(Qt.DisplayRole)
        self.close()


@register_binding_controller(ui_name="IconCommand",
                             klassname="DisplayIconCommand",
                             binding_type=SlotBinding,
                             priority=-10)
class DisplayIconCommand(BaseBindingController):
    # The scene model class for this controller
    model = Instance(DisplayIconCommandModel, args=())
    # Internal traits
    _button = Instance(QToolButton)
    _icon = Instance(QIcon)

    def create_widget(self, parent):
        # The ToolButton is affected with bounded actions, thus we need to
        # create another (parent) widget where we can bind the actions.
        widget = QWidget(parent)
        layout = QHBoxLayout(widget)

        self._button = DynamicToolButton(parent)
        self._button.clicked.connect(self.execute_action)
        layout.addWidget(self._button)

        change_icon = QAction("Change Icon", widget)
        change_icon.triggered.connect(self._change_icon)
        widget.addAction(change_icon)

        icon_name = self.model.icon_name
        self._icon = BUTTON_ICONS.get(icon_name, NO_SELECTION)
        self._button.setIcon(self._icon)

        return widget

    def execute_action(self):
        """Execute the action on the command proxy"""
        self.proxy.execute()

    def _change_icon(self):
        """Select the icon for the icon command widget"""
        dialog = IconSelectionDialog(parent=self.widget)
        dialog.exec()

        self._icon = dialog.icon
        self._button.setIcon(self._icon)
        self.model.icon_name = dialog.name

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self._button.setEnabled(enable)

    def setEnabled(self, enable):
        """Reimplemented to account for access level changes"""
        self._button.setEnabled(enable)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from enum import Enum
from functools import total_ordering
from heapq import heappop, heappush

from qtpy.QtCore import QObject, QRunnable, QThreadPool, QTimer, Signal, Slot


class TaskMediator(QObject):
    """A task mediator to launch finished signal"""
    finished = Signal(object)


class TaskResult:
    __slots__ = ("data", )

    def __init__(self, data=None):
        self.data = data

    def result(self):
        if isinstance(self.data, Exception):
            raise self.data
        return self.data

    def exception(self):
        if isinstance(self.data, Exception):
            return self.data
        return None


class Task(QRunnable):
    """A task that can be run the global instance of `QThreadPool`"""

    def __init__(self, obj, *args):
        super().__init__()
        self.obj = obj
        self.args = args
        self.signaller = None

    def run(self):
        """Reimplemented method of `QRunnable`"""
        try:
            data = self.obj(*self.args)
        except Exception as exc:
            data = exc
        if self.signaller is not None:
            result = TaskResult(data)
            self.signaller.finished.emit(result)
            self.signaller.deleteLater()

    def connect(self, callback):
        self.signaller = TaskMediator()
        self.signaller.finished.connect(callback)


def background(obj, *args, callback=None):
    """Run a function in the background in the global threadpool

    :returns: Returns task (QRunnable)
    """

    task = Task(obj, *args)
    if callback is not None:
        task.connect(callback)
    QThreadPool.globalInstance().start(task)
    return task


@total_ordering
class Priority(Enum):
    NETWORK = 0
    BIG_DATA = 1
    BACKGROUND = 2

    def __lt__(self, other):
        if isinstance(other, Priority):
            return self.value < other.value
        else:
            return NotImplemented


def executeLater(task, priority):
    """append a task to the queue of the ones to be executed"""
    global counter
    heappush(queue, (priority, counter, task))
    counter += 1
    timer.start()


MAX_ITEM_PROCESSING = 5
queue = []
counter = 0
timer = None


@Slot()
def timeout():
    """execute up to ``MAX_ITEM_PROCESSING`` tasks if existing

    If big data is in the stack, we cut there!
    """
    task_counter = MAX_ITEM_PROCESSING
    while queue and task_counter > 0:
        prio, _, task = heappop(queue)
        task()
        if prio.value > 0:
            # We found big data and break!
            break
        task_counter -= 1

    if not queue:
        timer.stop()


def create_background_timer():
    """Manually create the background `timer` after a QApplication has
    been created"""
    global timer
    timer = QTimer()
    timer.setInterval(0)
    timer.timeout.connect(timeout)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QColor, QFont, QIcon, QPixmap
from qtpy.QtWidgets import QColorDialog, QDialog

from karabo.common.scenemodel.api import LabelModel
from karabogui.dialogs.font_dialog import FontDialog
from karabogui.fonts import get_alias_from_font, get_qfont
from karabogui.util import SignalBlocker

from .utils import get_dialog_ui

BUTTON_SIZE = 10


class TextDialog(QDialog):

    def __init__(self, label_model=None, alignment=False, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('textdialog.ui'), self)
        self.has_alignment = alignment
        if label_model is None:
            # NOTE: Fonts similar on all OS are Arial, Helvetica, sans-serif!
            self.label_model = LabelModel()
        else:
            self.label_model = label_model.clone_traits()
        self.text_font = get_qfont(self.label_model.font, adjust_size=False)
        # Save font string after corrections
        self.label_model.font = self.text_font.toString()
        self.leText.setText(self.label_model.text)

        self.set_text_font_button()
        self.set_text_color_button()
        self.set_text_background_button()
        self.set_alignment_combo()
        if self.label_model.frame_width > 0:
            self.cbFrameWidth.setChecked(True)
            self.sbFrameWidth.setValue(self.label_model.frame_width)

        self.cbBackground.setChecked(
            self.label_model.background != 'transparent')

    def set_alignment_combo(self):
        """Set the combobox according to the alignment"""
        if self.has_alignment:
            alignment = Qt.AlignmentFlag(self.label_model.alignh)
            if alignment == Qt.AlignLeft:
                self.cbAlignment.setCurrentIndex(0)
            elif alignment == Qt.AlignRight:
                self.cbAlignment.setCurrentIndex(1)
            elif alignment == Qt.AlignHCenter:
                self.cbAlignment.setCurrentIndex(2)
        else:
            # There is no alignment, we are centered
            with SignalBlocker(self.cbAlignment):
                self.cbAlignment.setCurrentIndex(2)
                self.cbAlignment.setEnabled(False)

    def set_text_font_button(self):
        """
        Update the  text and font family of the button. The font size
        should not be changed in order to keep the button size fixed.
        """
        qfont = QFont(self.text_font)
        qfont.setPointSize(BUTTON_SIZE)
        self.pbFont.setFont(qfont)
        family = get_alias_from_font(self.text_font.family())
        button_text = f"{family}, {self.text_font.pointSize()}pt"
        self.pbFont.setText(button_text)

    def set_text_color_button(self):
        pixmap = QPixmap(24, 16)
        pixmap.fill(QColor(self.label_model.foreground))
        self.pbTextColor.setIcon(QIcon(pixmap))

    def set_text_background_button(self):
        pixmap = QPixmap(24, 16)
        pixmap.fill(QColor(self.label_model.background))
        self.pbBackground.setIcon(QIcon(pixmap))

    @Slot(int)
    def on_cbBackground_stateChanged(self, state):
        if state != Qt.Checked:
            self.label_model.background = 'transparent'
            self.set_text_background_button()

    @Slot(int)
    def on_cbAlignment_currentIndexChanged(self, index):
        if index == 0:
            alignment = Qt.AlignLeft
        elif index == 1:
            alignment = Qt.AlignRight
        elif index == 2:
            alignment = Qt.AlignHCenter

        # Store the int cast of the enum flag
        self.label_model.alignh = int(alignment)

    @Slot(str)
    def on_leText_textChanged(self, text):
        self.label_model.text = text

    @Slot()
    def on_pbFont_clicked(self):
        dialog = FontDialog(self.text_font, parent=self)
        if dialog.exec() == QDialog.Accepted:
            self.text_font = dialog.qfont
            self.label_model.font = self.text_font.toString()
            self.set_text_font_button()

    @Slot()
    def on_pbTextColor_clicked(self):
        color = QColorDialog.getColor(QColor(self.label_model.foreground))
        if color.isValid():
            self.label_model.foreground = color.name()
            self.set_text_color_button()

    @Slot(bool)
    def on_cbFrameWidth_toggled(self, checked):
        self.sbFrameWidth.setEnabled(checked)
        if not checked:
            self.label_model.frame_width = 0

    @Slot(int)
    def on_sbFrameWidth_valueChanged(self, value):
        self.label_model.frame_width = value

    @Slot(bool)
    def on_cbBackground_toggled(self, checked):
        self.pbBackground.setEnabled(checked)
        if not checked:
            self.label_model.background = ''

    @Slot()
    def on_pbBackground_clicked(self):
        color = QColorDialog.getColor(QColor(self.label_model.background))
        if color.isValid():
            self.label_model.background = color.name()
            self.set_text_background_button()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QDialog

from karabo.common.api import walk_traits_object
from karabo.common.scenemodel.api import SceneModel, SceneTargetWindow
from karabogui.singletons.api import get_project_model

from .utils import get_dialog_ui


class SceneLinkDialog(QDialog):
    def __init__(self, model, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui('scenelink.ui'), self)

        self._selectedScene = 0
        self._sceneTargets = self._get_scenelink_targets()
        for scene in self._sceneTargets:
            self.lwScenes.addItem(scene)

        flags = Qt.MatchExactly | Qt.MatchCaseSensitive
        self._select_item(model.target, flags)

        self._selectedTargetWin = model.target_window
        radioButtons = {
            SceneTargetWindow.MainWindow: self.mainRadio,
            SceneTargetWindow.Dialog: self.dialogRadio
        }
        button = radioButtons.get(self._selectedTargetWin)
        if button:
            button.setChecked(True)

    def _get_scenelink_targets(self):
        project = get_project_model().root_model
        if project is None:
            return []

        collected = set()

        def visitor(obj):
            nonlocal collected
            if isinstance(obj, SceneModel):
                target = f"{obj.simple_name}:{obj.uuid}"
                collected.add(target)

        walk_traits_object(project, visitor)
        return sorted(collected)

    def _select_item(self, text, flags):
        items = self.lwScenes.findItems(text, flags)
        if items:
            # Select first result
            self.lwScenes.setCurrentItem(items[0])
        self._selectedScene = self.lwScenes.currentRow()

    @property
    def selectedScene(self):
        return self._sceneTargets[self._selectedScene]

    @property
    def selectedTargetWindow(self):
        return self._selectedTargetWin

    @Slot(int)
    def on_lwScenes_currentRowChanged(self, index):
        self._selectedScene = index

    @Slot(bool)
    def on_dialogRadio_clicked(self, checked=False):
        if checked:
            self._selectedTargetWin = SceneTargetWindow.Dialog

    @Slot(bool)
    def on_mainRadio_clicked(self, checked=False):
        if checked:
            self._selectedTargetWin = SceneTargetWindow.MainWindow

    @Slot(str)
    def on_leFilter_textChanged(self, text):
        self._select_item(text, Qt.MatchStartsWith)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.


from qtpy.QtCore import QObject, Slot

from karabogui.testing import click_button

from ..switch import SwitchButton


class Mediator(QObject):
    state = False

    @Slot(bool)
    def receive(self, state):
        self.state = state


def test_switch_button(gui_app):
    mediator = Mediator()
    button = SwitchButton()
    button.toggled.connect(mediator.receive)
    click_button(button.button)
    assert mediator.state is True
    click_button(button.switch_frame)
    assert mediator.state is False



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from karabo.native import Configurable, VectorFloat
from karabogui.graph.plots.dialogs.data_analysis import DataAnalysisDialog
from karabogui.testing import get_class_property_proxy

CONFIG = {'half_samples': 6000, 'roi_items': [], 'roi_tool': 0, 'offset': 0.0,
          'step': 1.0, 'x_grid': True, 'y_grid': True, 'title': '',
          'background': 'transparent', 'x_label': '',
          'y_label': 'vectorProperty', 'x_units': '', 'y_units': '',
          'x_autorange': True, 'y_autorange': True, 'x_log': False,
          'y_log': False, 'x_invert': False, 'y_invert': False, 'x_min': 0.0,
          'x_max': 0.0, 'y_min': 0.0, 'y_max': 0.0, 'layout_data': None}


class Object(Configurable):
    prop = VectorFloat(defaultValue=[1.1, 2.0, 3.2, 4.5])
    prop2 = VectorFloat()
    baseline = VectorFloat()


def test_fetch_data(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    value = [1.0, 2.0, 3.0, 4.0]
    proxy.value = value

    dialog = DataAnalysisDialog(
        proxies=[proxy], config=CONFIG, parent=None)

    # Initial values
    combo = dialog.fit_options_combobox
    items = [combo.itemText(i) for i in range(combo.count())]
    expected = ["Gaussian", "Sech Square", "Step function (CDF)",
                "Linear function"]
    assert items == expected
    assert not dialog.fit_on_update_checkbox.isChecked()
    x, y = dialog.data_curve.getData()
    assert all(y == np.array(value))
    assert dialog.result_label.text() == ""
    assert list(dialog.x_values) == list(range(len(proxy.value)))
    assert_array_equal(dialog.y_values, proxy.value)

    # Update from the parent plot.
    value = [10, 20, 30]
    proxy.value = value
    dialog.update_data()
    x, y = dialog.data_curve.getData()
    assert all(y == np.array(value))


def test_multiple_proxies(gui_app, mocker):
    """
    Test dialog when the plot has multiple proxies.
    """
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    value = [1.0, 2.0, 3.0, 4.0]
    proxy.value = value

    additional_proxy = get_class_property_proxy(schema, "prop2")
    additional_proxy.value = [0.1, 0.2, 0.3, 0.4]

    dialog = DataAnalysisDialog(
        proxies=[proxy, additional_proxy], config=CONFIG, parent=None)
    combo_box = dialog.property_comboBox
    assert combo_box.currentText() == proxy.key
    assert combo_box.itemText(0) == proxy.key
    assert combo_box.itemText(1) == additional_proxy.key

    # When combobox item is changed, the plot should get updated.
    method = mocker.patch.object(dialog, "update_data")
    assert method.call_count == 0
    combo_box.setCurrentIndex(1)
    assert dialog.proxy == combo_box.currentData()
    assert dialog.proxy == additional_proxy
    assert method.call_count == 1


@pytest.mark.filterwarnings("ignore::scipy.optimize.OptimizeWarning")
def test_sub_region_roi(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    value = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
    proxy.value = value

    additional_proxy = get_class_property_proxy(schema, "prop2")
    additional_proxy.value = [0.1, 0.2, 0.3, 0.4]

    dialog = DataAnalysisDialog(
        proxies=[proxy, additional_proxy], config=CONFIG, parent=None)
    # Before fitting
    assert not dialog.show_line_roi_button.isChecked()
    assert dialog.left_line
    assert dialog.right_line
    assert not dialog.left_line.isVisible()
    assert not dialog.right_line.isVisible()

    assert dialog.fit_curve.getData() == (None, None)
    assert dialog._min_width == 1.0

    # Fit with no line roi defined.
    dialog.fit()
    _, y_values = dialog.fit_curve.getData()
    assert len(y_values) == 7

    # With line roi
    dialog.show_line_roi_button.click()
    assert dialog.show_line_roi_button.isChecked()
    assert dialog.left_line.isVisible()
    assert dialog.right_line.isVisible()
    assert dialog.left_line.pos().x() == 0
    assert dialog.right_line.pos().x() == 6.0

    # With whole plot selected
    dialog.fit()
    _, y_values = dialog.fit_curve.getData()
    assert len(y_values) == 6

    # With only a region selected
    dialog.left_line.setPos(1.0)
    dialog.right_line.setPos(5.0)
    dialog.fit()
    _, y_values = dialog.fit_curve.getData()
    assert len(y_values) == 4

    # Change property should update the lines position
    dialog.property_comboBox.setCurrentIndex(1)
    assert dialog.property_comboBox.currentText() == "prop2"
    assert dialog.left_line.pos().x() == 0
    assert dialog.right_line.pos().x() == 3.0
    assert dialog.left_pos_line_edit.text() == "0"
    assert dialog.right_pos_line_edit.text() == "3"


def test_sub_region_value_type(gui_app):
    """ Test that no traceback with Linear fit with selection"""
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    value = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
    proxy.value = value

    dialog = DataAnalysisDialog(
        proxies=[proxy], config=CONFIG, parent=None)
    fit_option = dialog.fit_options_combobox
    fit_option.setCurrentIndex(3)
    assert fit_option.currentText() == "Linear function"
    assert not dialog.show_line_roi_button.isChecked()
    dialog.fit_button.click()

    dialog.show_line_roi_button.setChecked(True)
    dialog.fit_button.click()


def test_baseline_proxy(gui_app):
    """Make sure the X-axis values are from baseline_proxy"""
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    value = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
    proxy.value = value

    baseline = get_class_property_proxy(schema, "baseline")
    baseline.value = [70, 60, 50, 40, 30, 20, 10]

    dialog = DataAnalysisDialog(
        proxies=[proxy], config=CONFIG, baseline_proxy=baseline)
    assert_array_equal(dialog.x_values, baseline.value)
    assert_array_equal(dialog.y_values, proxy.value)

    # Make sure the config options are affected only when generating
    # baseline and not with baseline_proxy.
    config = CONFIG.copy()
    config["step"] = 1.5
    config["offset"] = 3.0
    dialog = DataAnalysisDialog(
        proxies=[proxy], config=CONFIG, baseline_proxy=baseline)
    assert_array_equal(dialog.x_values, baseline.value)
    assert_array_equal(dialog.y_values, proxy.value)

    dialog = DataAnalysisDialog(
        proxies=[proxy], config=config, baseline_proxy=None)
    expected = np.array([3.0, 4.5, 6.0, 7.5, 9.0, 10.5, 12])
    assert_array_equal(dialog.x_values, expected)

    assert np.isclose(dialog._min_width, 1.425, rtol=0.005)


def test_sech_sequred(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    value = [1.0, 2.0, 3.0, 5.0, 7.0, 9.0, 10.0, 8.0, 6.0, 4.0]
    proxy.value = value

    dialog = DataAnalysisDialog(
        proxies=[proxy], config=CONFIG, parent=None)
    fit_option = dialog.fit_options_combobox
    fit_option.setCurrentIndex(1)
    assert fit_option.currentText() == "Sech Square"
    dialog.fit()
    fit_x, fit_y = dialog.fit_curve.getData()
    expected = np.array([0.98434883, 1.76325783, 3.04367143, 4.930586,
                        7.21354721, 9.10750318, 9.56312261, 8.27814066,
                        6.0563046, 3.91234989])
    np.isclose(expected, fit_y)


def test_auto_update(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    value = [1.0, 2.0, 3.0, 5.0, 7.0, 9.0, 10.0, 8.0, 6.0, 4.0]
    proxy.value = value
    dialog = DataAnalysisDialog(proxies=[proxy], config=CONFIG, parent=None)
    x, y = dialog.data_curve.getData()
    assert_array_equal(y, value)

    # Plot should not be updated on proxy update.
    dialog.proxy.binding.config_update = True
    new_value = np.array([60, 50, 40, 30, 20, 10])
    proxy.value = new_value
    x, y = dialog.data_curve.getData()
    with np.testing.assert_raises(AssertionError):
        assert_array_equal(y, new_value)

    # With auto_update checked, plot should be updated on proxy update.
    dialog.auto_update.setChecked(True)
    dialog.proxy.binding.config_update = True
    dialog.proxy.value = new_value
    x, y = dialog.data_curve.getData()
    assert_array_equal(y, new_value)



#############################################################################
# Author: <raul.costa@xfel.eu>
# Created on August 26, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy import uic
from qtpy.QtCore import QSize, Qt, Slot
from qtpy.QtWidgets import (
    QAbstractItemView, QDialog, QDialogButtonBox, QTableWidgetItem)

from karabogui import messagebox
from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.singletons.api import get_config, get_db_conn
from karabogui.util import SignalBlocker, get_spin_widget, utc_to_local

from .utils import get_dialog_ui

# Minimum size for a device id substring to be considered searchable.
MIN_NAME_SIZE = 4


class LoadProjectWithDialog(QDialog):

    def __init__(self, domain=None, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui("load_project_with_item.ui"), self)
        self.setWindowTitle("Find and Load Project")
        self.domain = domain
        self.text_name_part.setPlaceholderText(
            f"Enter name part (at least {MIN_NAME_SIZE} "
            "characters)")
        self.button_box.button(QDialogButtonBox.Ok).setText("Load")
        flags = Qt.WindowCloseButtonHint
        self.setWindowFlags(self.windowFlags() | flags)

        self.spin_widget = get_spin_widget(icon="wait-black",
                                           scaled_size=QSize(16, 16))
        self.spin_widget.setVisible(False)
        self.bottom_horizontal_layout.insertWidget(0, self.spin_widget)

        # Subscribes for a set of broadcast events.
        self.event_map = {
            KaraboEvent.ProjectDomainsList: self._event_domain_list,
            KaraboEvent.ProjectFindWithItem: self._event_find_with_item
        }
        register_for_broadcasts(self.event_map)

        self.setup_table()
        self.connect_widgets_events()

        self.loading_domains = False
        self.finding_projects = False
        self.db_conn = get_db_conn()
        self.db_conn.ignore_local_cache = True
        self.load_domains()

    def done(self, result):
        """ Override ``QDialog`` virtual slot.
        Unsubscribes from broadcast events.
        """
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    # -----------------------------------------------------------------------
    # Dialog values accessors

    def selected_project_id(self):
        project_id = None
        if len(self.table_projects.selectedItems()) == 3:
            # The project name is the first item and the uuid is its
            # associated data.
            project_name = self.table_projects.selectedItems()[0]
            project_id = project_name.data(Qt.UserRole)
        return project_id

    def selected_domain(self):
        domain = None
        if len(self.combo_domain.currentText()) > 0:
            domain = self.combo_domain.currentText()
        return domain

    @property
    def simple_name(self):
        item = self.table_projects.selectedItems()[0]
        name = item.data(Qt.DisplayRole)
        return name

    # -----------------------------------------------------------------------
    # Karabo Event Handlers

    def _event_domain_list(self, data):
        self._domains_updated(data.get("items", []))

    def _event_find_with_item(self, data):
        if self.finding_projects:
            # The results are for the search currently being expected.
            # A long-running search could have been triggered from a
            # previous instance of the dialog.
            self._projects_updated(data)

    def _domains_updated(self, domains):
        domain_to_select = self.domain
        with SignalBlocker(self.combo_domain):
            self.combo_domain.clear()
            self.combo_domain.addItems(sorted(domains))
            # Sets the selected domain to the one passed to the dialog's
            # initializer (if there was one) or sets the selected domain to
            # the current topic (if it is among the returned domains).
            domain_index = -1
            if domain_to_select:
                domain_index = self.combo_domain.findText(domain_to_select)
            if domain_index == -1:
                current_topic = get_config()["broker_topic"]
                topic_index = self.combo_domain.findText(current_topic)
                if topic_index >= 0:
                    # The topic is among the domains
                    self.combo_domain.setCurrentIndex(topic_index)
            else:
                # The previously selected domain has been found
                self.combo_domain.setCurrentIndex(domain_index)
        self.loading_domains = False
        self.update_dialog_state()

    def _projects_updated(self, data):
        self.finding_projects = False
        if data["error"] is not None:
            self.update_dialog_state()
            messagebox.show_error("Find projects failed!",
                                  details=data["error"])
        else:
            self.populate_projects_table(data["items"])
            self.update_dialog_state()

    # -----------------------------------------------------------------------
    # Widgets Event Handlers

    def connect_widgets_events(self):
        self.button_find_projects.clicked.connect(self.find_projects)
        self.combo_domain.currentIndexChanged.connect(self.refresh_search)

        self.text_name_part.textEdited.connect(self.text_edited)
        self.text_name_part.returnPressed.connect(self.refresh_search)
        self.combo_item_type.currentIndexChanged.connect(self.refresh_search)

        # Double-clicking a project with a device should load it.
        self.table_projects.doubleClicked.connect(self.accept)
        self.table_projects.itemSelectionChanged.connect(
            self.update_dialog_state)

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    @Slot()
    def refresh_search(self):
        if len(self.text_name_part.text()) >= MIN_NAME_SIZE:
            self.find_projects()

    @Slot(str)
    def text_edited(self, text):
        if not self.finding_projects:
            enabled = len(text) >= MIN_NAME_SIZE
            if enabled:
                msg = ("Click <b>Find Projects</b> "
                       f"to search for <b>{text}</b>.")
            else:
                msg = (f"Enter name part with at least <b>{MIN_NAME_SIZE} </b>"
                       "characters.")
            self.label_status.setText(msg)
            self.button_find_projects.setEnabled(enabled)

    # -----------------------------------------------------------------------
    # Data Fetching

    def load_domains(self):
        self.loading_domains = True
        self.db_conn.get_available_domains()
        self.update_dialog_state()

    def find_projects(self):
        self.finding_projects = True
        self.clear_table()
        self.update_dialog_state()
        search_domain = self.combo_domain.currentText()
        name_part = self.text_name_part.text()

        if self.combo_item_type.currentText() == "Device":
            self.db_conn.get_projects_with_device(
                search_domain, name_part)
        elif self.combo_item_type.currentText() == "Macro":
            self.db_conn.get_projects_with_macro(
                search_domain, name_part)
        elif self.combo_item_type.currentText() == "Server":
            self.db_conn.get_projects_with_server(
                search_domain, name_part)

    # -----------------------------------------------------------------------
    # UI Setup and Updating

    def setup_table(self):
        cols = ["Project", "Last Modified", "Matching"]
        self.table_projects.verticalHeader().setVisible(False)
        self.table_projects.setColumnCount(len(cols))
        self.table_projects.setHorizontalHeaderLabels(cols)
        self.table_projects.horizontalHeader().setStretchLastSection(True)
        self.table_projects.setAlternatingRowColors(True)
        self.table_projects.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_projects.setSelectionMode(QAbstractItemView.SingleSelection)
        self.table_projects.setEditTriggers(QAbstractItemView.NoEditTriggers)

    def clear_table(self):
        self.table_projects.clearContents()
        self.table_projects.setRowCount(0)

    def populate_projects_table(self, data):
        if len(data) == 0:
            return
        self.table_projects.setSortingEnabled(False)
        self.table_projects.setUpdatesEnabled(False)
        self.table_projects.setRowCount(len(data))
        for row, rec in enumerate(data):
            project_name = QTableWidgetItem(rec["project_name"])
            project_name.setData(Qt.UserRole, rec["uuid"])
            self.table_projects.setItem(row, 0, project_name)
            self.table_projects.setItem(
                row, 1, QTableWidgetItem(utc_to_local(rec["date"])))
            items = rec.get("items", [])
            items = ", ".join(sorted(items))
            table_item = QTableWidgetItem(items)
            table_item.setToolTip(items)
            self.table_projects.setItem(row, 2, table_item)

        self.table_projects.setUpdatesEnabled(True)
        self.table_projects.setSortingEnabled(True)
        # The last column is not resized to avoid conflict with its stretch
        # attribute. Resizing it may result in uneeded space for a vertical
        # scroll bar.
        self.table_projects.resizeColumnToContents(0)
        self.table_projects.resizeColumnToContents(1)

    def update_dialog_state(self):
        loading_data = self.loading_domains or self.finding_projects
        name_part = self.text_name_part.text()

        self.combo_domain.setEnabled(not loading_data)
        self.combo_item_type.setEnabled(not loading_data)
        enabled = (not loading_data and len(name_part) >= MIN_NAME_SIZE)
        self.button_find_projects.setEnabled(enabled)
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(
            len(self.table_projects.selectedItems()) > 0)

        search_domain = self.combo_domain.currentText()
        if search_domain and name_part and not loading_data:
            if self.table_projects.rowCount() == 0:
                self.label_status.setText(
                    f"No project with {name_part} in "
                    f"{search_domain}")
            elif self.table_projects.rowCount() == 1:
                self.label_status.setText(
                    f"1 project with {name_part} in "
                    f"{search_domain}")
            else:
                self.label_status.setText(
                    f"{self.table_projects.rowCount()} projects with "
                    f"{name_part} in {search_domain}")
        elif self.loading_domains:
            self.label_status.setText("Loading domains ...")
        elif self.finding_projects:
            self.label_status.setText("Finding projects ...")
        else:
            self.label_status.setText("")

        self.spin_widget.setVisible(loading_data)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtWidgets import QAction, QInputDialog
from traits.api import Instance, WeakRef

from karabo.common.scenemodel.api import (
    VectorBarGraphModel, build_model_config)
from karabogui.binding.api import NDArrayBinding, VectorNumberBinding
from karabogui.controllers.api import (
    BaseBindingController, get_array_data, register_binding_controller)
from karabogui.graph.common.const import ACTION_ITEMS
from karabogui.graph.plots.api import (
    KaraboPlotView, VectorBarGraphPlot, generate_down_sample, get_view_range)

MAX_WIDTH = 100
MAX_BARS = 3000

# Do not allow to set logarithmic scale on X-axis.This is to avoid overriding
# methods from upstream as pyqtgraph doesn't support log values on axes, yet
actions = [action for action in ACTION_ITEMS if action != 'x_log']


@register_binding_controller(
    ui_name='Vector Bar Graph',
    klassname='VectorBarGraph',
    binding_type=(NDArrayBinding, VectorNumberBinding),
    can_show_nothing=False)
class DisplayVectorBarGraph(BaseBindingController):
    """The BarGraph controller for display of pulse data in a histogram format
    """
    model = Instance(VectorBarGraphModel, args=())
    _plot = WeakRef(VectorBarGraphPlot)

    def create_widget(self, parent):
        widget = KaraboPlotView(parent=parent, actions=actions)
        widget.add_cross_target()
        widget.add_toolbar()
        widget.stateChanged.connect(self._change_model)
        self._plot = widget.add_bar_item()
        self._plot.set_width(self.model.bar_width)

        widget.restore(build_model_config(self.model))

        width_action = QAction("Bar Width", widget)
        width_action.triggered.connect(self.configure_bar_width)
        widget.addAction(width_action)

        return widget

    def value_update(self, proxy):
        value, _ = get_array_data(proxy, default=[])
        if not len(value):
            self._plot.setData([], [])
            return

        rect = get_view_range(self._plot)
        x, y = generate_down_sample(value, threshold=MAX_BARS, rect=rect)
        self._plot.setData(x, y)

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**content)

    def configure_bar_width(self):
        bar_width, ok = QInputDialog.getDouble(
            self.widget, 'Bar Width', 'Bar Width:',
            self.model.bar_width, 0.1, MAX_WIDTH)
        if ok:
            self.model.bar_width = bar_width
            self._plot.set_width(bar_width)
